"""The metrics module implements independent and collections of Prometheus metrics that can be used in Argo."""

from typing import List, Optional, Union

from hera.shared import BaseMixin
from hera.workflows.models import (
    Amount as _ModelAmount,
    Counter as _ModelCounter,
    Gauge as _ModelGauge,
    Histogram as _ModelHistogram,
    MetricLabel as _ModelMetricLabel,
    Prometheus as _ModelPrometheus,
)

Label = _ModelMetricLabel


class _BaseMetric(BaseMixin):
    """Base metric wrapper around `hera.workflows.models.Prometheus`."""

    name: str
    help: str
    labels: Optional[Union[Label, List[Label]]] = None
    when: Optional[str] = None

    def _build_labels(self) -> Optional[List[_ModelMetricLabel]]:
        if not self.labels:
            return None
        if isinstance(self.labels, Label):
            return [self.labels]
        return self.labels

    def _build_metric(self) -> _ModelPrometheus:
        raise NotImplementedError


class Counter(_BaseMetric):
    """Counter metric component used to count specific events based on the given value.

    Notes:
        See [https://argoproj.github.io/argo-workflows/metrics/#grafana-dashboard-for-argo-controller-metrics](https://argoproj.github.io/argo-workflows/metrics/#grafana-dashboard-for-argo-controller-metrics)
    """

    value: str

    def _build_metric(self) -> _ModelPrometheus:
        return _ModelPrometheus(
            counter=_ModelCounter(value=self.value),
            gauge=None,
            help=self.help,
            histogram=None,
            labels=self._build_labels(),
            name=self.name,
            when=self.when,
        )


class Gauge(_BaseMetric, _ModelGauge):
    """Gauge metric component used to record intervals based on the given value.

    Notes:
        See [https://argoproj.github.io/argo-workflows/metrics/#grafana-dashboard-for-argo-controller-metrics](https://argoproj.github.io/argo-workflows/metrics/#grafana-dashboard-for-argo-controller-metrics)
    """

    realtime: bool
    value: str

    def _build_metric(self) -> _ModelPrometheus:
        return _ModelPrometheus(
            counter=None,
            gauge=_ModelGauge(realtime=self.realtime, value=self.value),
            help=self.help,
            histogram=None,
            labels=self._build_labels(),
            name=self.name,
            when=self.when,
        )


class Histogram(_BaseMetric):
    """Histogram metric that records the value at the specified bucket intervals.

    Notes:
        See [https://argoproj.github.io/argo-workflows/metrics/#grafana-dashboard-for-argo-controller-metrics](https://argoproj.github.io/argo-workflows/metrics/#grafana-dashboard-for-argo-controller-metrics)
    """

    buckets: List[Union[float, _ModelAmount]]  # type: ignore
    value: str

    def _build_buckets(self) -> List[_ModelAmount]:
        return [_ModelAmount(__root__=bucket) if isinstance(bucket, float) else bucket for bucket in self.buckets]

    def _build_metric(self) -> _ModelPrometheus:
        return _ModelPrometheus(
            counter=None,
            gauge=None,
            help=self.help,
            histogram=_ModelHistogram(
                buckets=self._build_buckets(),
                value=self.value,
            ),
            labels=self._build_labels(),
            name=self.name,
            when=self.when,
        )


class Metric(_BaseMetric):
    """Prometheus metric that can be used at the workflow or task/template level.

    Notes:
        See [https://argoproj.github.io/argo-workflows/metrics/#grafana-dashboard-for-argo-controller-metrics](https://argoproj.github.io/argo-workflows/metrics/#grafana-dashboard-for-argo-controller-metrics)
    """

    counter: Optional[Counter] = None
    gauge: Optional[Gauge] = None
    histogram: Optional[Histogram] = None

    def _build_metric(self) -> _ModelPrometheus:
        return _ModelPrometheus(
            counter=_ModelCounter(value=self.counter.value) if self.counter else None,
            gauge=_ModelGauge(realtime=self.gauge.realtime, value=self.gauge.value) if self.gauge else None,
            help=self.help,
            histogram=_ModelHistogram(
                buckets=self.histogram._build_buckets(),
                value=self.histogram.value,
            )
            if self.histogram
            else None,
            labels=self._build_labels(),
            name=self.name,
            when=self.when,
        )


class Metrics(BaseMixin):
    """A collection of Prometheus metrics.

    Notes:
        See [https://argoproj.github.io/argo-workflows/metrics/#grafana-dashboard-for-argo-controller-metrics](https://argoproj.github.io/argo-workflows/metrics/#grafana-dashboard-for-argo-controller-metrics)
    """

    metrics: List[Metric]

    def _build_metrics(self) -> List[_ModelPrometheus]:
        return [metric._build_metric() for metric in self.metrics]


__all__ = [
    "Label",
    "Counter",
    "Gauge",
    "Histogram",
    "Metric",
    "Metrics",
]



"""The HTTP template module provides functionality and objects required for executing HTTP calls in workflows."""

from typing import List, Optional

from hera.workflows._meta_mixins import CallableTemplateMixin
from hera.workflows._mixins import IOMixin, TemplateMixin
from hera.workflows.models import (
    HTTP as _ModelHTTP,
    HTTPBodySource,
    Template as _ModelTemplate,
    V1alpha1HTTPHeader as HTTPHeader,
)


class HTTP(TemplateMixin, IOMixin, CallableTemplateMixin):
    """`HTTP` is an implementation of the HTTP template that supports executing HTTP actions in a step/task."""

    url: str
    body: Optional[str] = None
    body_from: Optional[HTTPBodySource] = None
    headers: Optional[List[HTTPHeader]] = None
    insecure_skip_verify: Optional[bool] = None
    method: Optional[str] = None
    success_condition: Optional[str] = None
    timeout_seconds: Optional[int] = None

    def _build_http_template(self) -> _ModelHTTP:
        """Builds the generated HTTP sub-template."""
        return _ModelHTTP(
            url=self.url,
            body=self.body,
            body_from=self.body_from,
            headers=self.headers,
            insecure_skip_verify=self.insecure_skip_verify,
            method=self.method,
            success_condition=self.success_condition,
            timeout_seconds=self.timeout_seconds,
        )

    def _build_template(self) -> _ModelTemplate:
        """Builds the HTTP generated `Template`."""
        return _ModelTemplate(
            active_deadline_seconds=self.active_deadline_seconds,
            affinity=self.affinity,
            archive_location=self.archive_location,
            automount_service_account_token=self.automount_service_account_token,
            executor=self.executor,
            fail_fast=self.fail_fast,
            host_aliases=self.host_aliases,
            http=self._build_http_template(),
            init_containers=self._build_init_containers(),
            memoize=self.memoize,
            metadata=self._build_metadata(),
            inputs=self._build_inputs(),
            outputs=self._build_outputs(),
            name=self.name,
            node_selector=self.node_selector,
            plugin=self.plugin,
            priority=self.priority,
            priority_class_name=self.priority_class_name,
            retry_strategy=self.retry_strategy,
            scheduler_name=self.scheduler_name,
            security_context=self.pod_security_context,
            service_account_name=self.service_account_name,
            sidecars=self._build_sidecars(),
            synchronization=self.synchronization,
            timeout=self.timeout,
            tolerations=self.tolerations,
        )


__all__ = ["HTTP"]



"""[DO NOT EDIT MANUALLY] Auto-generated model classes.

Auto-generated by Hera via `make events-models`.
OpenAPI spec URL: https://raw.githubusercontent.com/argoproj/argo-workflows/v3.6.2/api/openapi-spec/swagger.json
"""

from hera.events.models.eventsource import (
    CreateEventSourceRequest,
    EventSourceDeletedResponse,
    EventSourceWatchEvent,
    LogEntry as EventsourceLogEntry,
    UpdateEventSourceRequest,
)
from hera.events.models.io.argoproj.events.v1alpha1 import (
    Amount,
    AMQPConsumeConfig,
    AMQPEventSource,
    AMQPExchangeDeclareConfig,
    AMQPQueueBindConfig,
    AMQPQueueDeclareConfig,
    ArgoWorkflowTrigger,
    ArtifactLocation,
    AWSLambdaTrigger,
    AzureEventHubsTrigger,
    AzureEventsHubEventSource,
    AzureQueueStorageEventSource,
    AzureServiceBusEventSource,
    AzureServiceBusTrigger,
    Backoff,
    BasicAuth,
    BitbucketAuth,
    BitbucketBasicAuth,
    BitbucketEventSource,
    BitbucketRepository,
    BitbucketServerEventSource,
    BitbucketServerRepository,
    CalendarEventSource,
    CatchupConfiguration,
    Condition,
    ConditionsResetByTime,
    ConditionsResetCriteria,
    ConfigMapPersistence,
    CustomTrigger,
    DataFilter,
    EmailTrigger,
    EmitterEventSource,
    EventContext,
    EventDependency,
    EventDependencyFilter,
    EventDependencyTransformer,
    EventPersistence,
    EventSource,
    EventSourceFilter,
    EventSourceList,
    EventSourceSpec,
    EventSourceStatus,
    ExprFilter,
    FileArtifact,
    FileEventSource,
    GenericEventSource,
    GerritEventSource,
    GitArtifact,
    GitCreds,
    GithubAppCreds,
    GithubEventSource,
    GitlabEventSource,
    GitRemoteConfig,
    HDFSEventSource,
    HTTPTrigger,
    Int64OrString,
    K8SResourcePolicy,
    KafkaConsumerGroup,
    KafkaEventSource,
    KafkaTrigger,
    LogTrigger,
    Metadata,
    MQTTEventSource,
    NATSAuth,
    NATSEventsSource,
    NATSTrigger,
    NSQEventSource,
    OpenWhiskTrigger,
    OwnedRepositories,
    PayloadField,
    PubSubEventSource,
    PulsarEventSource,
    PulsarTrigger,
    RateLimit,
    RedisEventSource,
    RedisStreamEventSource,
    Resource,
    ResourceEventSource,
    ResourceFilter,
    S3Artifact,
    S3Bucket,
    S3Filter,
    SASLConfig,
    SchemaRegistryConfig,
    SecureHeader,
    Selector,
    Sensor,
    SensorList,
    SensorSpec,
    SensorStatus,
    Service,
    SFTPEventSource,
    SlackEventSource,
    SlackSender,
    SlackThread,
    SlackTrigger,
    SNSEventSource,
    SQSEventSource,
    StandardK8STrigger,
    Status,
    StatusPolicy,
    StorageGridEventSource,
    StorageGridFilter,
    StripeEventSource,
    Template,
    TimeFilter,
    TLSConfig,
    Trigger,
    TriggerParameter,
    TriggerParameterSource,
    TriggerPolicy,
    TriggerTemplate,
    URLArtifact,
    ValueFromSource,
    WatchPathConfig,
    WebhookContext,
    WebhookEventSource,
)
from hera.events.models.io.argoproj.workflow.v1alpha1 import (
    Event,
    EventResponse,
    GetUserInfoResponse,
    InfoResponse,
    Item,
    Version,
)
from hera.events.models.io.k8s.api.core.v1 import ImagePullPolicy
from hera.events.models.sensor import (
    CreateSensorRequest,
    DeleteSensorResponse,
    LogEntry as SensorLogEntry,
    SensorWatchEvent,
    UpdateSensorRequest,
)

__all__ = [
    "AMQPConsumeConfig",
    "AMQPEventSource",
    "AMQPExchangeDeclareConfig",
    "AMQPQueueBindConfig",
    "AMQPQueueDeclareConfig",
    "AWSLambdaTrigger",
    "Amount",
    "ArgoWorkflowTrigger",
    "ArtifactLocation",
    "AzureEventHubsTrigger",
    "AzureEventsHubEventSource",
    "AzureQueueStorageEventSource",
    "AzureServiceBusEventSource",
    "AzureServiceBusTrigger",
    "Backoff",
    "BasicAuth",
    "BitbucketAuth",
    "BitbucketBasicAuth",
    "BitbucketEventSource",
    "BitbucketRepository",
    "BitbucketServerEventSource",
    "BitbucketServerRepository",
    "CalendarEventSource",
    "CatchupConfiguration",
    "Condition",
    "ConditionsResetByTime",
    "ConditionsResetCriteria",
    "ConfigMapPersistence",
    "CreateEventSourceRequest",
    "CreateSensorRequest",
    "CustomTrigger",
    "DataFilter",
    "DeleteSensorResponse",
    "EmailTrigger",
    "EmitterEventSource",
    "Event",
    "EventContext",
    "EventDependency",
    "EventDependencyFilter",
    "EventDependencyTransformer",
    "EventPersistence",
    "EventResponse",
    "EventSource",
    "EventSourceDeletedResponse",
    "EventSourceFilter",
    "EventSourceList",
    "EventSourceSpec",
    "EventSourceStatus",
    "EventSourceWatchEvent",
    "EventsourceLogEntry",
    "ExprFilter",
    "FileArtifact",
    "FileEventSource",
    "GenericEventSource",
    "GerritEventSource",
    "GetUserInfoResponse",
    "GitArtifact",
    "GitCreds",
    "GitRemoteConfig",
    "GithubAppCreds",
    "GithubEventSource",
    "GitlabEventSource",
    "HDFSEventSource",
    "HTTPTrigger",
    "ImagePullPolicy",
    "InfoResponse",
    "Int64OrString",
    "Item",
    "K8SResourcePolicy",
    "KafkaConsumerGroup",
    "KafkaEventSource",
    "KafkaTrigger",
    "LogTrigger",
    "MQTTEventSource",
    "Metadata",
    "NATSAuth",
    "NATSEventsSource",
    "NATSTrigger",
    "NSQEventSource",
    "OpenWhiskTrigger",
    "OwnedRepositories",
    "PayloadField",
    "PubSubEventSource",
    "PulsarEventSource",
    "PulsarTrigger",
    "RateLimit",
    "RedisEventSource",
    "RedisStreamEventSource",
    "Resource",
    "ResourceEventSource",
    "ResourceFilter",
    "S3Artifact",
    "S3Bucket",
    "S3Filter",
    "SASLConfig",
    "SFTPEventSource",
    "SNSEventSource",
    "SQSEventSource",
    "SchemaRegistryConfig",
    "SecureHeader",
    "Selector",
    "Sensor",
    "SensorList",
    "SensorLogEntry",
    "SensorSpec",
    "SensorStatus",
    "SensorWatchEvent",
    "Service",
    "SlackEventSource",
    "SlackSender",
    "SlackThread",
    "SlackTrigger",
    "StandardK8STrigger",
    "Status",
    "StatusPolicy",
    "StorageGridEventSource",
    "StorageGridFilter",
    "StripeEventSource",
    "TLSConfig",
    "Template",
    "TimeFilter",
    "Trigger",
    "TriggerParameter",
    "TriggerParameterSource",
    "TriggerPolicy",
    "TriggerTemplate",
    "URLArtifact",
    "UpdateEventSourceRequest",
    "UpdateSensorRequest",
    "ValueFromSource",
    "Version",
    "WatchPathConfig",
    "WebhookContext",
    "WebhookEventSource",
]



"""Provides Hera implementations of Argo's container node for usage in Workflow tasks/steps."""

from __future__ import annotations

from typing import Any, List, Optional, Union

from hera.workflows._meta_mixins import CallableTemplateMixin, ContextMixin
from hera.workflows._mixins import (
    ContainerMixin,
    EnvIOMixin,
    EnvMixin,
    ResourceMixin,
    SubNodeMixin,
    TemplateMixin,
    VolumeMountMixin,
)
from hera.workflows.exceptions import InvalidType
from hera.workflows.models import (
    ContainerNode as _ModelContainerNode,
    ContainerSetRetryStrategy,
    ContainerSetTemplate as _ModelContainerSetTemplate,
    Lifecycle,
    SecurityContext,
    Template as _ModelTemplate,
)


class ContainerNode(ContainerMixin, VolumeMountMixin, ResourceMixin, EnvMixin, SubNodeMixin):
    """A regular container that can be used as part of a `hera.workflows.ContainerSet`.

    See Also:
        [Container Set Template](https://argoproj.github.io/argo-workflows/container-set-template/)
    """

    name: str
    args: Optional[List[str]] = None
    command: Optional[List[str]] = None
    dependencies: Optional[List[str]] = None
    lifecycle: Optional[Lifecycle] = None
    security_context: Optional[SecurityContext] = None
    working_dir: Optional[str] = None

    def next(self, other: ContainerNode) -> ContainerNode:
        """Sets the given container as a dependency of this container and returns the given container.

        Examples:
            >>> from hera.workflows import ContainerNode
            >>> a, b = ContainerNode(name="a"), ContainerNode(name="b")
            >>> a.next(b)
            >>> b.dependencies
            ['a']
        """
        assert issubclass(other.__class__, ContainerNode)
        if other.dependencies is None:
            other.dependencies = [self.name]
        else:
            other.dependencies.append(self.name)
        other.dependencies = sorted(list(set(other.dependencies)))
        return other

    def __rrshift__(self, other: List[ContainerNode]) -> ContainerNode:
        """Sets `self` as a dependent of the given list of other `hera.workflows.ContainerNode`.

        Practically, the `__rrshift__` allows us to express statements such as `[a, b, c] >> d`, where `d` is `self.`

        Examples:
            >>> from hera.workflows import ContainerNode
            >>> a, b, c = ContainerNode(name="a"), ContainerNode(name="b"), ContainerNode(name="c")
            >>> [a, b] >> c
            >>> c.dependencies
            ['a', 'b']
        """
        assert isinstance(other, list), f"Unknown type {type(other)} specified using reverse right bitshift operator"
        for o in other:
            o.next(self)
        return self

    def __rshift__(
        self, other: Union[ContainerNode, List[ContainerNode]]
    ) -> Union[ContainerNode, List[ContainerNode]]:
        """Sets the given container as a dependency of this container and returns the given container.

        Examples:
            >>> from hera.workflows import ContainerNode
            >>> a, b = ContainerNode(name="a"), ContainerNode(name="b")
            >>> a >> b
            >>> b.dependencies
            ['a']
        """
        if isinstance(other, ContainerNode):
            return self.next(other)
        elif isinstance(other, list):
            for o in other:
                assert isinstance(o, ContainerNode), (
                    f"Unknown list item type {type(o)} specified using right bitshift operator `>>`"
                )
                self.next(o)
            return other
        raise ValueError(f"Unknown type {type(other)} provided to `__rshift__`")

    def _build_container_node(self) -> _ModelContainerNode:
        """Builds the generated `ContainerNode`."""
        return _ModelContainerNode(
            args=self.args,
            command=self.command,
            dependencies=self.dependencies,
            env=self._build_env(),
            env_from=self._build_env_from(),
            image=self.image,
            image_pull_policy=self._build_image_pull_policy(),
            lifecycle=self.lifecycle,
            liveness_probe=self.liveness_probe,
            name=self.name,
            ports=self.ports,
            readiness_probe=self.readiness_probe,
            resize_policy=self.resize_policy,
            resources=self._build_resources(),
            restart_policy=self.restart_policy,
            security_context=self.security_context,
            startup_probe=self.startup_probe,
            stdin=self.stdin,
            stdin_once=self.stdin_once,
            termination_message_path=self.termination_message_path,
            termination_message_policy=self.termination_message_policy,
            tty=self.tty,
            volume_devices=self.volume_devices,
            volume_mounts=self._build_volume_mounts(),
            working_dir=self.working_dir,
        )


class ContainerSet(
    EnvIOMixin,
    ContainerMixin,
    TemplateMixin,
    CallableTemplateMixin,
    VolumeMountMixin,
    ContextMixin,
):
    """`ContainerSet` is the implementation of a set of containers that can be run in parallel on Kubernetes.

    The containers are run within the same pod.

    Examples:
        >>> with ContainerSet(...) as cs:
        >>>     ContainerNode(...)
        >>>     ContainerNode(...)
    """

    containers: List[Union[ContainerNode, _ModelContainerNode]] = []
    container_set_retry_strategy: Optional[ContainerSetRetryStrategy] = None

    def _add_sub(self, node: Any):
        if not isinstance(node, ContainerNode):
            raise InvalidType(type(node))

        self.containers.append(node)

    def _build_container_set(self) -> _ModelContainerSetTemplate:
        """Builds the generated `ContainerSetTemplate`."""
        containers = [c._build_container_node() if isinstance(c, ContainerNode) else c for c in self.containers]
        return _ModelContainerSetTemplate(
            containers=containers,
            retry_strategy=self.container_set_retry_strategy,
            volume_mounts=self.volume_mounts,
        )

    def _build_template(self) -> _ModelTemplate:
        """Builds the generated `Template` representation of the container set."""
        return _ModelTemplate(
            active_deadline_seconds=self.active_deadline_seconds,
            affinity=self.affinity,
            archive_location=self.archive_location,
            automount_service_account_token=self.automount_service_account_token,
            container_set=self._build_container_set(),
            daemon=self.daemon,
            executor=self.executor,
            fail_fast=self.fail_fast,
            host_aliases=self.host_aliases,
            init_containers=self._build_init_containers(),
            inputs=self._build_inputs(),
            memoize=self.memoize,
            metadata=self._build_metadata(),
            metrics=self._build_metrics(),
            name=self.name,
            node_selector=self.node_selector,
            outputs=self._build_outputs(),
            plugin=self.plugin,
            pod_spec_patch=self.pod_spec_patch,
            priority=self.priority,
            priority_class_name=self.priority_class_name,
            retry_strategy=self.retry_strategy,
            scheduler_name=self.scheduler_name,
            security_context=self.pod_security_context,
            service_account_name=self.service_account_name,
            sidecars=self._build_sidecars(),
            synchronization=self.synchronization,
            timeout=self.timeout,
            tolerations=self.tolerations,
            volumes=self._build_volumes(),
        )


__all__ = ["ContainerNode", "ContainerSet"]



"""Pydantic V1 input/output models for the Hera runner."""

from typing import Any

from hera.shared._pydantic import BaseModel
from hera.workflows.io._io_mixins import InputMixin, OutputMixin


class Input(InputMixin, BaseModel):
    """Input model usable by the Hera Runner.

    Input is a Pydantic model which users can create a subclass of. When a subclass
    of Input is used as a function parameter type, Hera will take the fields of the
    user's subclass to create template input parameters and artifacts. See the example
    for the script_pydantic_io experimental feature.
    """


class Output(OutputMixin, BaseModel):
    """Output model usable by the Hera Runner.

    Output is a Pydantic model which users can create a subclass of. When a subclass
    of Output is used as a function return type, Hera will take the fields of the
    user's subclass to create template output parameters and artifacts. See the example
    for the script_pydantic_io experimental feature.
    """

    exit_code: int = 0
    result: Any = None



"""Hera makes Python code easy to orchestrate on Argo Workflows through native Python integrations.

The main goal of Hera is to make the Argo ecosystem accessible by simplifying
workflow construction and submission. Hera presents an intuitive Python interface
to the underlying API of Argo, with custom classes making use of context managers
and callables, empowering users to focus on their own executable payloads rather
than workflow setup.
"""

from hera._version import version

__version__ = version
__version_info__ = version.split(".")



from types import ModuleType
from typing import Optional

_yaml: Optional[ModuleType] = None
try:
    import yaml

    _yaml = yaml
except ImportError:
    _yaml = None
else:

    def str_presenter(dumper, data):
        """Configures yaml for dumping multiline strings.

        Ref: https://github.com/yaml/pyyaml/issues/240
        """
        if data.count("\n") > 0:  # check for multiline string
            return dumper.represent_scalar("tag:yaml.org,2002:str", data, style="|")
        return dumper.represent_scalar("tag:yaml.org,2002:str", data)

    _yaml.add_representer(str, str_presenter)
    _yaml.representer.SafeRepresenter.add_representer(str, str_presenter)


def dump(*args, **kwargs) -> str:
    """Builds the Workflow as an Argo schema Workflow object and returns it as yaml string."""
    if not _yaml:
        raise ImportError("`PyYAML` is not installed. Install `hera[yaml]` to bring in the extra dependency")

    # Set some default options if not provided by the user
    kwargs.setdefault("default_flow_style", False)
    kwargs.setdefault("sort_keys", False)
    return _yaml.dump(*args, **kwargs)



"""The template_set module provides the TemplateSet class."""

from typing import Any, List, Union

from hera.workflows._meta_mixins import TemplateDecoratorFuncsMixin
from hera.workflows.exceptions import InvalidType
from hera.workflows.models import (
    Template as _ModelTemplate,
)
from hera.workflows.protocol import Templatable


class TemplateSet(
    TemplateDecoratorFuncsMixin,
):
    """The TemplateSet class can be used to arrange templates across modules in a package."""

    templates: List[Union[_ModelTemplate, Templatable]] = []

    def _add_sub(self, node: Any):
        """Adds the given node (expected to satisfy the `Templatable` protocol) to the context."""
        if not isinstance(node, (Templatable, _ModelTemplate)):
            raise InvalidType(type(node))
        self.templates.append(node)


__all__ = ["TemplateSet"]



"""A serialization module that contains utilities for serializing any values passed to the Argo server via Hera."""

import json
from json import JSONEncoder
from typing import Any, Optional

from pydantic import BaseModel

# NOTE: Use the original BaseModel in order to support serializing user-defined models,
# for hera internal models, we still need to support v1 base models.
from hera.shared._pydantic import _PYDANTIC_VERSION

if _PYDANTIC_VERSION == 2:
    from pydantic.v1 import BaseModel as V1BaseModel  # type: ignore
else:
    V1BaseModel = None  # type: ignore

MISSING = object()
"""`MISSING` is a placeholder that indicates field value nullity.

When the user of a Hera object sets the field of an object specifically to `None`, Hera needs to distinguish between
default nullity/None and user-provided `None` on, say, something like the `source` of `Script`.
"""


class PydanticEncoder(JSONEncoder):
    """Default serializer of Hera objects."""

    def default(self, o: Any):
        """Return the default representation of the given object."""
        # Note that these are slightly different outputs b/w v1 and v2
        # v1 will give the actual python object whereas v2 will serialize it into
        # a json compatible format.
        if _PYDANTIC_VERSION == 1:
            if isinstance(o, BaseModel):
                return o.dict(by_alias=True)
        else:
            if isinstance(o, BaseModel):
                return o.model_dump(by_alias=True, mode="json")  # type: ignore
            if isinstance(o, V1BaseModel):
                return o.dict(by_alias=True)
        return super().default(o)


def serialize(value: Any) -> Optional[str]:
    """Serialize the given value.

    If the value is `MISSING` then a proper `None` is returned. Since strings are "serialized" already, they are simply
    returned. Everything else is JSON encoded and returned.
    """
    if value == MISSING:
        return None
    elif isinstance(value, str):
        return value
    else:
        return json.dumps(value, cls=PydanticEncoder)  # None serialized as `null`



# generated by datamodel-codegen:
#   filename:  argo-workflows-3.6.2.json

from __future__ import annotations

from typing import Union

from hera.shared._pydantic import BaseModel


class IntOrString(BaseModel):
    __root__: Union[str, int]



"""The parameter module provides the Parameter class.

See https://argoproj.github.io/argo-workflows/walk-through/parameters/
for a tutorial on Parameters.
"""

from __future__ import annotations

from typing import Any, Optional

from hera.shared._pydantic import root_validator
from hera.shared.serialization import MISSING, serialize
from hera.workflows.models import Parameter as _ModelParameter


class Parameter(_ModelParameter):
    """A `Parameter` is used to pass values in and out of templates.

    They are to declare input and output parameters in the case of templates, and are used
    for Steps and Tasks to assign values.
    """

    name: Optional[str] = None  # type: ignore

    output: Optional[bool] = False
    """used to specify parameter as an output in function signature annotations"""

    def _check_name(self):
        if not self.name:
            raise ValueError("name cannot be `None` or empty when used")

    # `MISSING` is the default value so that `Parameter` serialization understands the difference between a
    # missing value and a value of `None`, as set by a user. With this, when something sets a value of `None` it is
    # taken as a proper `None`. By comparison, if a user does not set a value, it is taken as `MISSING` and therefore
    # not serialized. This happens because the values if turned into an _actual_ `None` by `serialize` and therefore
    # Pydantic will not include it in the YAML that is passed to Argo
    value: Optional[Any] = MISSING
    default: Optional[Any] = MISSING

    @root_validator(pre=True, allow_reuse=True)
    def _check_values(cls, values):
        if values.get("value") is not None and values.get("value_from") is not None:
            raise ValueError("Cannot specify both `value` and `value_from` when instantiating `Parameter`")

        values["value"] = serialize(values.get("value", MISSING))
        values["default"] = serialize(values.get("default", MISSING))
        if values.get("enum", []):
            # We don't need to set "enum" in values to "MISSING" if there are no values
            # as it's a list of values. The values themselves should be serialized.
            values["enum"] = [serialize(v) for v in values.get("enum")]

        return values

    @classmethod
    def _get_input_attributes(cls):
        """Return the attributes used for input parameter annotations."""
        return ["enum", "description", "default", "name", "value", "value_from"]

    def __str__(self):
        """Represent the parameter as a string by pointing to its value.

        This is useful in situations where we want to concatenate string values such as
        Task.args=["echo", wf.get_parameter("message")].
        """
        if self.value is None:
            raise ValueError("Cannot represent `Parameter` as string as `value` is not set")
        return self.value

    @classmethod
    def from_model(cls, model: _ModelParameter) -> Parameter:
        """Creates a `Parameter` from a `Parameter` model."""
        return cls(**model.dict())

    def with_name(self, name: str) -> Parameter:
        """Returns a copy of the parameter with the name set to the value."""
        p = self.copy(deep=True)
        p.name = name
        return p

    def as_input(self) -> _ModelParameter:
        """Assembles the parameter for use as an input of a template."""
        self._check_name()
        assert self.name
        return _ModelParameter(
            name=self.name,
            description=self.description,
            default=self.default,
            enum=self.enum,
            value=self.value,
            value_from=self.value_from,
        )

    def as_argument(self) -> _ModelParameter:
        """Assembles the parameter for use as an argument of a step or a task."""
        # Setting a default value when used as an argument is a no-op so we exclude it as it would get overwritten by
        # `value` or `value_from` (one of which is required)
        # Overwrite ref: https://github.com/argoproj/argo-workflows/blob/781675ddcf6f1138d697cb9c71dae484daa0548b/workflow/common/util.go#L126-L139
        # One of value/value_from required ref: https://github.com/argoproj/argo-workflows/blob/ab178bb0b36a5ce34b4c1302cf4855879a0e8cf5/workflow/validate/validate.go#L794-L798
        self._check_name()
        assert self.name
        return _ModelParameter(
            name=self.name,
            global_name=self.global_name,
            description=self.description,
            enum=self.enum,
            value=self.value,
            value_from=self.value_from,
        )

    def as_output(self) -> _ModelParameter:
        """Assembles the parameter for use as an output of a template."""
        # Only `value` and `value_from` are valid here
        # see https://github.com/argoproj/argo-workflows/blob/e3254eca115c9dd358e55d16c6a3d41403c29cae/workflow/validate/validate.go#L1067
        self._check_name()
        assert self.name
        return _ModelParameter(
            name=self.name,
            global_name=self.global_name,
            description=self.description,
            value=self.value,
            value_from=self.value_from,
        )


__all__ = ["Parameter"]



# generated by datamodel-codegen:
#   filename:  argo-workflows-3.6.2.json

from __future__ import annotations

from typing import Annotated

from hera.shared._pydantic import BaseModel, Field


class Quantity(BaseModel):
    __root__: Annotated[
        str,
        Field(
            description=(
                "Quantity is a fixed-point representation of a number. It provides"
                " convenient marshaling/unmarshaling in JSON and YAML, in addition to"
                " String() and AsInt64() accessors.\n\nThe serialization format"
                " is:\n\n``` <quantity>        ::= <signedNumber><suffix>\n\n\t(Note"
                ' that <suffix> may be empty, from the "" case in'
                " <decimalSI>.)\n\n<digit>           ::= 0 | 1 | ... | 9 <digits>      "
                "    ::= <digit> | <digit><digits> <number>          ::= <digits> |"
                ' <digits>.<digits> | <digits>. | .<digits> <sign>            ::= "+" |'
                ' "-" <signedNumber>    ::= <number> | <sign><number> <suffix>         '
                " ::= <binarySI> | <decimalExponent> | <decimalSI> <binarySI>       "
                " ::= Ki | Mi | Gi | Ti | Pi | Ei\n\n\t(International System of units;"
                " See: http://physics.nist.gov/cuu/Units/binary.html)\n\n<decimalSI>   "
                '    ::= m | "" | k | M | G | T | P | E\n\n\t(Note that 1024 = 1Ki but'
                " 1000 = 1k; I didn't choose the capitalization.)\n\n<decimalExponent>"
                ' ::= "e" <signedNumber> | "E" <signedNumber> ```\n\nNo matter which of'
                " the three exponent forms is used, no quantity may represent a number"
                " greater than 2^63-1 in magnitude, nor may it have more than 3 decimal"
                " places. Numbers larger or more precise will be capped or rounded up."
                " (E.g.: 0.1m will rounded up to 1m.) This may be extended in the"
                " future if we require larger or smaller quantities.\n\nWhen a Quantity"
                " is parsed from a string, it will remember the type of suffix it had,"
                " and will use the same type again when it is serialized.\n\nBefore"
                ' serializing, Quantity will be put in "canonical form". This means'
                " that Exponent/suffix will be adjusted up or down (with a"
                " corresponding increase or decrease in Mantissa) such that:\n\n- No"
                " precision is lost - No fractional digits will be emitted - The"
                " exponent (or suffix) is as large as possible.\n\nThe sign will be"
                " omitted unless the number is negative.\n\nExamples:\n\n- 1.5 will be"
                ' serialized as "1500m" - 1.5Gi will be serialized as "1536Mi"\n\nNote'
                " that the quantity will NEVER be internally represented by a floating"
                " point number. That is the whole point of this"
                " exercise.\n\nNon-canonical values will still parse as long as they"
                " are well formed, but will be re-emitted in their canonical form. (So"
                " always use canonical form, or don't diff.)\n\nThis format is"
                " intended to make it difficult to use these numbers without writing"
                " some sort of special handling code in the hopes that that will cause"
                " implementors to also use a fixed point implementation."
            )
        ),
    ]



"""The shared module of Hera provides control over global configurations, hooks, and some base mixins."""

from hera.shared._global_config import BaseMixin, GlobalConfig, global_config, register_pre_build_hook

__all__ = ["global_config", "register_pre_build_hook", "GlobalConfig", "BaseMixin"]



"""Entrypoint for running hera as a CLI."""

import importlib
import sys

from hera._version import version


def main(argv=None):
    """Entrypoint for running hera as a CLI."""
    try:
        importlib.import_module("cappa")
    except ModuleNotFoundError as e:
        raise ModuleNotFoundError(
            "Use of the `hera` CLI tool requires installing the 'cli' extra, `pip install hera[cli]`."
        ) from e

    import cappa
    import rich

    from hera._cli.base import Hera

    rich.print(
        "[yellow bold]warning: The `hera` CLI is a work-in-progress, subject to change at any time![/yellow bold]",
        file=sys.stderr,
    )

    cappa.invoke(Hera, argv=argv, version=version)


if __name__ == "__main__":  # pragma: no cover
    main()



# generated by datamodel-codegen:
#   filename:  argo-workflows-3.6.2.json

from __future__ import annotations

from typing import Optional

from hera.shared._pydantic import BaseModel


class Any(BaseModel):
    type_url: Optional[str] = None
    value: Optional[str] = None



"""The resource module provides functionality for creating K8s resources via workflows inside task/steps."""

from typing import List, Optional, Union

from hera.workflows._meta_mixins import CallableTemplateMixin
from hera.workflows._mixins import IOMixin, SubNodeMixin, TemplateMixin
from hera.workflows.cron_workflow import CronWorkflow
from hera.workflows.models import (
    ManifestFrom,
    ResourceTemplate as _ModelResourceTemplate,
    Template as _ModelTemplate,
)
from hera.workflows.workflow import Workflow
from hera.workflows.workflow_template import WorkflowTemplate


class Resource(CallableTemplateMixin, TemplateMixin, SubNodeMixin, IOMixin):
    """`Resource` is a representation of a K8s resource that can be created by Argo.

    The resource is a callable step that can be invoked in a DAG/Workflow. The resource can create any K8s resource,
    such as other workflows, workflow templates, daemons, etc, as specified by the `manifest` field of the resource.
    The manifest field is a canonical YAML that is submitted to K8s by Argo. Note that the manifest is a union of
    multiple types. The manifest can be a string, in which case it is assume to be YAML. Otherwise, if it's a Hera
    objects, it is automatically converted to the corresponding YAML representation.
    """

    action: str
    failure_condition: Optional[str] = None
    flags: Optional[List[str]] = None
    manifest: Optional[Union[str, Workflow, CronWorkflow, WorkflowTemplate]] = None
    manifest_from: Optional[ManifestFrom] = None
    merge_strategy: Optional[str] = None
    set_owner_reference: Optional[bool] = None
    success_condition: Optional[str] = None

    def _build_manifest(self) -> Optional[str]:
        if isinstance(self.manifest, (Workflow, CronWorkflow, WorkflowTemplate)):
            # hack to appease raw yaml string comparison
            return self.manifest.to_yaml().replace("'{{", "{{").replace("}}'", "}}")
        return self.manifest

    def _build_resource_template(self) -> _ModelResourceTemplate:
        return _ModelResourceTemplate(
            action=self.action,
            failure_condition=self.failure_condition,
            flags=self.flags,
            manifest=self._build_manifest(),
            manifest_from=self.manifest_from,
            merge_strategy=self.merge_strategy,
            set_owner_reference=self.set_owner_reference,
            success_condition=self.success_condition,
        )

    def _build_template(self) -> _ModelTemplate:
        return _ModelTemplate(
            active_deadline_seconds=self.active_deadline_seconds,
            affinity=self.affinity,
            archive_location=self.archive_location,
            automount_service_account_token=self.automount_service_account_token,
            daemon=self.daemon,
            executor=self.executor,
            fail_fast=self.fail_fast,
            host_aliases=self.host_aliases,
            init_containers=self._build_init_containers(),
            inputs=self._build_inputs(),
            memoize=self.memoize,
            metadata=self._build_metadata(),
            metrics=self._build_metrics(),
            name=self.name,
            node_selector=self.node_selector,
            outputs=self._build_outputs(),
            parallelism=self.parallelism,
            plugin=self.plugin,
            pod_spec_patch=self.pod_spec_patch,
            priority=self.priority,
            priority_class_name=self.priority_class_name,
            resource=self._build_resource_template(),
            retry_strategy=self.retry_strategy,
            scheduler_name=self.scheduler_name,
            security_context=self.pod_security_context,
            service_account_name=self.service_account_name,
            sidecars=self._build_sidecars(),
            synchronization=self.synchronization,
            timeout=self.timeout,
            tolerations=self.tolerations,
        )


__all__ = ["Resource"]



"""Utility functions for converting binary and decimal units to float values."""

import re

_decimal_multipliers = {
    "m": 1e-3,
    "k": 1e3,
    "M": 1e6,
    "G": 1e9,
    "T": 1e12,
    "P": 1e15,
    "E": 1e18,
}

_binary_multipliers = {
    "Ki": 2**10,
    "Mi": 2**20,
    "Gi": 2**30,
    "Ti": 2**40,
    "Pi": 2**50,
    "Ei": 2**60,
}


def _convert_decimal_units(value: str) -> float:
    """Converts the given decimal units to a float value. If no unit is given, the value is multiplied by 1.

    Args:
        value (str): The value to convert the decimal units of. The supported units are ['m', 'k', 'M', 'G', 'T', 'P', 'E'].
        Note that the units are case sensitive.

    Raises:
        ValueError: When the identified unit is not a supported one.

    Returns:
        float: Float value of the given decimal units.
    """
    pattern = r"^\s*([+-]?\d+(?:\.\d+)?)([mkMGTPE]?)\s*$"
    matches = re.match(pattern, value)

    if matches:
        value, unit = matches.groups()
        return float(value) * _decimal_multipliers.get(unit, 1)
    else:
        raise ValueError(
            f"Invalid decimal units for input: {value}. Supported units are ['m', 'k', 'M', 'G', 'T', 'P', 'E']."
        )


def _convert_binary_units(value: str) -> float:
    """Converts the given binary units to a float value. If no unit is given, the value is multiplied by 1.

    Args:
        value (str): The value to convert the binary unit of. The supported units are ['Ki', 'Mi', 'Gi', 'Ti', 'Pi', 'Ei'].
        Note that the units are case sensitive.

    Raises:
        ValueError: When the identified unit is not a supported one.

    Returns:
        float: Float value of the given binary units.
    """
    pattern = r"^\s*([+-]?\d+(?:\.\d+)?)([KMGTPE]i)?\s*$"
    matches = re.match(pattern, value)

    if matches:
        value, unit = matches.groups()
        return float(value) * _binary_multipliers.get(unit, 1)
    else:
        raise ValueError(
            f"Invalid binary units for input: {value}. Supported units are ['Ki', 'Mi', 'Gi', 'Ti', 'Pi', 'Ei']."
        )


convert_memory_units = _convert_binary_units
convert_storage_units = _convert_binary_units
convert_cpu_units = _convert_decimal_units



# generated by datamodel-codegen:
#   filename:  argo-workflows-3.6.2.json

from __future__ import annotations

from typing import Annotated, Optional

from hera.shared._pydantic import BaseModel, Field

from .io.argoproj.events import v1alpha1
from .io.k8s.apimachinery.pkg.apis.meta import v1


class DeleteSensorResponse(BaseModel):
    pass


class LogEntry(BaseModel):
    dependency_name: Annotated[
        Optional[str],
        Field(alias="dependencyName", title="optional - trigger dependency name"),
    ] = None
    event_context: Annotated[
        Optional[str],
        Field(alias="eventContext", title="optional - Cloud Event context"),
    ] = None
    level: Optional[str] = None
    msg: Optional[str] = None
    namespace: Optional[str] = None
    sensor_name: Annotated[Optional[str], Field(alias="sensorName")] = None
    time: Optional[v1.Time] = None
    trigger_name: Annotated[Optional[str], Field(alias="triggerName", title="optional - any trigger name")] = None


class CreateSensorRequest(BaseModel):
    create_options: Annotated[Optional[v1.CreateOptions], Field(alias="createOptions")] = None
    namespace: Optional[str] = None
    sensor: Optional[v1alpha1.Sensor] = None


class SensorWatchEvent(BaseModel):
    object: Optional[v1alpha1.Sensor] = None
    type: Optional[str] = None


class UpdateSensorRequest(BaseModel):
    name: Optional[str] = None
    namespace: Optional[str] = None
    sensor: Optional[v1alpha1.Sensor] = None



"""Interact with the REST service."""

# [DO NOT MODIFY] Auto-generated by `hera.scripts.service.py`
from typing import Optional, Tuple, cast
from urllib.parse import urljoin

import requests

from hera.events.models import (
    CreateEventSourceRequest,
    CreateSensorRequest,
    DeleteSensorResponse,
    Event,
    EventResponse,
    EventSource,
    EventSourceDeletedResponse,
    EventSourceList,
    EventsourceLogEntry,
    EventSourceWatchEvent,
    GetUserInfoResponse,
    InfoResponse,
    Item,
    Sensor,
    SensorList,
    SensorLogEntry,
    SensorWatchEvent,
    UpdateEventSourceRequest,
    UpdateSensorRequest,
    Version,
)
from hera.exceptions import exception_from_server_response
from hera.shared import global_config


def valid_host_scheme(host: str) -> bool:
    """Validates the the given `host` starts with either `http` or `https`."""
    return host.startswith("http://") or host.startswith("https://")


class EventsService:
    """The core events service for interacting with the Argo server."""

    def __init__(
        self,
        host: Optional[str] = None,
        verify_ssl: Optional[bool] = None,
        token: Optional[str] = None,
        client_certs: Optional[Tuple[str, str]] = None,
        namespace: Optional[str] = None,
        session: Optional[requests.Session] = None,
    ) -> None:
        """Events service constructor."""
        self.host = cast(str, host or global_config.host)
        self.verify_ssl = verify_ssl if verify_ssl is not None else global_config.verify_ssl
        self.client_certs = client_certs if client_certs is not None else global_config.client_certs

        # some users reported in https://github.com/argoproj-labs/hera/issues/1016 that it can be a bit awkward for
        # Hera to assume a `Bearer` prefix on behalf of users. Some might pass it and some might not. Therefore, Hera
        # only prefixes the token with `Bearer ` if it's not already specified and lets the uses specify it otherwise.
        # Note that the `Bearer` token can be specified through the global configuration as well. In order to deliver
        # a fix on Hera V5 without introducing breaking changes, we have to support both
        global_config_token = global_config.token  # call only once because it can be a user specified function!

        def format_token(t):
            parts = t.strip().split()
            if len(parts) == 1:
                return "Bearer " + t
            return t

        if token:
            self.token: Optional[str] = format_token(token)
        elif global_config_token:
            self.token = format_token(global_config_token)
        else:
            self.token = None

        self.session = session or requests.Session()

        self.namespace = namespace or global_config.namespace

    def _request(self, method, **kwargs):
        """Make a request using the session if enabled."""
        return self.session.request(method, **kwargs)

    def close(self):
        """Close the service session."""
        self.session.close()

    def __enter__(self):
        """Open the service - session doesn't need to be opened."""
        return self

    def __exit__(self, *_):
        """Close the service."""
        self.close()

    def list_event_sources(
        self,
        namespace: Optional[str] = None,
        label_selector: Optional[str] = None,
        field_selector: Optional[str] = None,
        watch: Optional[bool] = None,
        allow_watch_bookmarks: Optional[bool] = None,
        resource_version: Optional[str] = None,
        resource_version_match: Optional[str] = None,
        timeout_seconds: Optional[str] = None,
        limit: Optional[str] = None,
        continue_: Optional[str] = None,
        send_initial_events: Optional[bool] = None,
    ) -> EventSourceList:
        """API documentation."""
        assert valid_host_scheme(self.host), "The host scheme is required for service usage"
        resp = self._request(
            method="get",
            url=urljoin(self.host, "api/v1/event-sources/{namespace}").format(
                namespace=namespace if namespace is not None else self.namespace
            ),
            params={
                "listOptions.labelSelector": label_selector,
                "listOptions.fieldSelector": field_selector,
                "listOptions.watch": watch,
                "listOptions.allowWatchBookmarks": allow_watch_bookmarks,
                "listOptions.resourceVersion": resource_version,
                "listOptions.resourceVersionMatch": resource_version_match,
                "listOptions.timeoutSeconds": timeout_seconds,
                "listOptions.limit": limit,
                "listOptions.continue": continue_,
                "listOptions.sendInitialEvents": send_initial_events,
            },
            headers={"Authorization": self.token},
            data=None,
            verify=self.verify_ssl,
            cert=self.client_certs,
        )

        if resp.ok:
            return EventSourceList(**resp.json())

        raise exception_from_server_response(resp)

    def create_event_source(self, req: CreateEventSourceRequest, namespace: Optional[str] = None) -> EventSource:
        """API documentation."""
        assert valid_host_scheme(self.host), "The host scheme is required for service usage"
        resp = self._request(
            method="post",
            url=urljoin(self.host, "api/v1/event-sources/{namespace}").format(
                namespace=namespace if namespace is not None else self.namespace
            ),
            params=None,
            headers={"Authorization": self.token, "Content-Type": "application/json"},
            data=req.json(exclude_none=True, by_alias=True, exclude_unset=True, exclude_defaults=True),
            verify=self.verify_ssl,
            cert=self.client_certs,
        )

        if resp.ok:
            return EventSource(**resp.json())

        raise exception_from_server_response(resp)

    def get_event_source(self, name: str, namespace: Optional[str] = None) -> EventSource:
        """API documentation."""
        assert valid_host_scheme(self.host), "The host scheme is required for service usage"
        resp = self._request(
            method="get",
            url=urljoin(self.host, "api/v1/event-sources/{namespace}/{name}").format(
                name=name, namespace=namespace if namespace is not None else self.namespace
            ),
            params=None,
            headers={"Authorization": self.token},
            data=None,
            verify=self.verify_ssl,
            cert=self.client_certs,
        )

        if resp.ok:
            return EventSource(**resp.json())

        raise exception_from_server_response(resp)

    def update_event_source(
        self, name: str, req: UpdateEventSourceRequest, namespace: Optional[str] = None
    ) -> EventSource:
        """API documentation."""
        assert valid_host_scheme(self.host), "The host scheme is required for service usage"
        resp = self._request(
            method="put",
            url=urljoin(self.host, "api/v1/event-sources/{namespace}/{name}").format(
                name=name, namespace=namespace if namespace is not None else self.namespace
            ),
            params=None,
            headers={"Authorization": self.token, "Content-Type": "application/json"},
            data=req.json(exclude_none=True, by_alias=True, exclude_unset=True, exclude_defaults=True),
            verify=self.verify_ssl,
            cert=self.client_certs,
        )

        if resp.ok:
            return EventSource(**resp.json())

        raise exception_from_server_response(resp)

    def delete_event_source(
        self,
        name: str,
        namespace: Optional[str] = None,
        grace_period_seconds: Optional[str] = None,
        uid: Optional[str] = None,
        resource_version: Optional[str] = None,
        orphan_dependents: Optional[bool] = None,
        propagation_policy: Optional[str] = None,
        dry_run: Optional[list] = None,
    ) -> EventSourceDeletedResponse:
        """API documentation."""
        assert valid_host_scheme(self.host), "The host scheme is required for service usage"
        resp = self._request(
            method="delete",
            url=urljoin(self.host, "api/v1/event-sources/{namespace}/{name}").format(
                name=name, namespace=namespace if namespace is not None else self.namespace
            ),
            params={
                "deleteOptions.gracePeriodSeconds": grace_period_seconds,
                "deleteOptions.preconditions.uid": uid,
                "deleteOptions.preconditions.resourceVersion": resource_version,
                "deleteOptions.orphanDependents": orphan_dependents,
                "deleteOptions.propagationPolicy": propagation_policy,
                "deleteOptions.dryRun": dry_run,
            },
            headers={"Authorization": self.token},
            data=None,
            verify=self.verify_ssl,
            cert=self.client_certs,
        )

        if resp.ok:
            return EventSourceDeletedResponse()

        raise exception_from_server_response(resp)

    def receive_event(self, discriminator: str, req: Item, namespace: Optional[str] = None) -> EventResponse:
        """API documentation."""
        assert valid_host_scheme(self.host), "The host scheme is required for service usage"
        resp = self._request(
            method="post",
            url=urljoin(self.host, "api/v1/events/{namespace}/{discriminator}").format(
                discriminator=discriminator, namespace=namespace if namespace is not None else self.namespace
            ),
            params=None,
            headers={"Authorization": self.token, "Content-Type": "application/json"},
            data=req.json(exclude_none=True, by_alias=True, exclude_unset=True, exclude_defaults=True),
            verify=self.verify_ssl,
            cert=self.client_certs,
        )

        if resp.ok:
            return EventResponse()

        raise exception_from_server_response(resp)

    def get_info(self) -> InfoResponse:
        """API documentation."""
        assert valid_host_scheme(self.host), "The host scheme is required for service usage"
        resp = self._request(
            method="get",
            url=urljoin(self.host, "api/v1/info"),
            params=None,
            headers={"Authorization": self.token},
            data=None,
            verify=self.verify_ssl,
            cert=self.client_certs,
        )

        if resp.ok:
            return InfoResponse(**resp.json())

        raise exception_from_server_response(resp)

    def list_sensors(
        self,
        namespace: Optional[str] = None,
        label_selector: Optional[str] = None,
        field_selector: Optional[str] = None,
        watch: Optional[bool] = None,
        allow_watch_bookmarks: Optional[bool] = None,
        resource_version: Optional[str] = None,
        resource_version_match: Optional[str] = None,
        timeout_seconds: Optional[str] = None,
        limit: Optional[str] = None,
        continue_: Optional[str] = None,
        send_initial_events: Optional[bool] = None,
    ) -> SensorList:
        """API documentation."""
        assert valid_host_scheme(self.host), "The host scheme is required for service usage"
        resp = self._request(
            method="get",
            url=urljoin(self.host, "api/v1/sensors/{namespace}").format(
                namespace=namespace if namespace is not None else self.namespace
            ),
            params={
                "listOptions.labelSelector": label_selector,
                "listOptions.fieldSelector": field_selector,
                "listOptions.watch": watch,
                "listOptions.allowWatchBookmarks": allow_watch_bookmarks,
                "listOptions.resourceVersion": resource_version,
                "listOptions.resourceVersionMatch": resource_version_match,
                "listOptions.timeoutSeconds": timeout_seconds,
                "listOptions.limit": limit,
                "listOptions.continue": continue_,
                "listOptions.sendInitialEvents": send_initial_events,
            },
            headers={"Authorization": self.token},
            data=None,
            verify=self.verify_ssl,
            cert=self.client_certs,
        )

        if resp.ok:
            return SensorList(**resp.json())

        raise exception_from_server_response(resp)

    def create_sensor(self, req: CreateSensorRequest, namespace: Optional[str] = None) -> Sensor:
        """API documentation."""
        assert valid_host_scheme(self.host), "The host scheme is required for service usage"
        resp = self._request(
            method="post",
            url=urljoin(self.host, "api/v1/sensors/{namespace}").format(
                namespace=namespace if namespace is not None else self.namespace
            ),
            params=None,
            headers={"Authorization": self.token, "Content-Type": "application/json"},
            data=req.json(exclude_none=True, by_alias=True, exclude_unset=True, exclude_defaults=True),
            verify=self.verify_ssl,
            cert=self.client_certs,
        )

        if resp.ok:
            return Sensor(**resp.json())

        raise exception_from_server_response(resp)

    def get_sensor(self, name: str, namespace: Optional[str] = None, resource_version: Optional[str] = None) -> Sensor:
        """API documentation."""
        assert valid_host_scheme(self.host), "The host scheme is required for service usage"
        resp = self._request(
            method="get",
            url=urljoin(self.host, "api/v1/sensors/{namespace}/{name}").format(
                name=name, namespace=namespace if namespace is not None else self.namespace
            ),
            params={"getOptions.resourceVersion": resource_version},
            headers={"Authorization": self.token},
            data=None,
            verify=self.verify_ssl,
            cert=self.client_certs,
        )

        if resp.ok:
            return Sensor(**resp.json())

        raise exception_from_server_response(resp)

    def update_sensor(self, name: str, req: UpdateSensorRequest, namespace: Optional[str] = None) -> Sensor:
        """API documentation."""
        assert valid_host_scheme(self.host), "The host scheme is required for service usage"
        resp = self._request(
            method="put",
            url=urljoin(self.host, "api/v1/sensors/{namespace}/{name}").format(
                name=name, namespace=namespace if namespace is not None else self.namespace
            ),
            params=None,
            headers={"Authorization": self.token, "Content-Type": "application/json"},
            data=req.json(exclude_none=True, by_alias=True, exclude_unset=True, exclude_defaults=True),
            verify=self.verify_ssl,
            cert=self.client_certs,
        )

        if resp.ok:
            return Sensor(**resp.json())

        raise exception_from_server_response(resp)

    def delete_sensor(
        self,
        name: str,
        namespace: Optional[str] = None,
        grace_period_seconds: Optional[str] = None,
        uid: Optional[str] = None,
        resource_version: Optional[str] = None,
        orphan_dependents: Optional[bool] = None,
        propagation_policy: Optional[str] = None,
        dry_run: Optional[list] = None,
    ) -> DeleteSensorResponse:
        """API documentation."""
        assert valid_host_scheme(self.host), "The host scheme is required for service usage"
        resp = self._request(
            method="delete",
            url=urljoin(self.host, "api/v1/sensors/{namespace}/{name}").format(
                name=name, namespace=namespace if namespace is not None else self.namespace
            ),
            params={
                "deleteOptions.gracePeriodSeconds": grace_period_seconds,
                "deleteOptions.preconditions.uid": uid,
                "deleteOptions.preconditions.resourceVersion": resource_version,
                "deleteOptions.orphanDependents": orphan_dependents,
                "deleteOptions.propagationPolicy": propagation_policy,
                "deleteOptions.dryRun": dry_run,
            },
            headers={"Authorization": self.token},
            data=None,
            verify=self.verify_ssl,
            cert=self.client_certs,
        )

        if resp.ok:
            return DeleteSensorResponse()

        raise exception_from_server_response(resp)

    def watch_event_sources(
        self,
        namespace: Optional[str] = None,
        label_selector: Optional[str] = None,
        field_selector: Optional[str] = None,
        watch: Optional[bool] = None,
        allow_watch_bookmarks: Optional[bool] = None,
        resource_version: Optional[str] = None,
        resource_version_match: Optional[str] = None,
        timeout_seconds: Optional[str] = None,
        limit: Optional[str] = None,
        continue_: Optional[str] = None,
        send_initial_events: Optional[bool] = None,
    ) -> EventSourceWatchEvent:
        """API documentation."""
        assert valid_host_scheme(self.host), "The host scheme is required for service usage"
        resp = self._request(
            method="get",
            url=urljoin(self.host, "api/v1/stream/event-sources/{namespace}").format(
                namespace=namespace if namespace is not None else self.namespace
            ),
            params={
                "listOptions.labelSelector": label_selector,
                "listOptions.fieldSelector": field_selector,
                "listOptions.watch": watch,
                "listOptions.allowWatchBookmarks": allow_watch_bookmarks,
                "listOptions.resourceVersion": resource_version,
                "listOptions.resourceVersionMatch": resource_version_match,
                "listOptions.timeoutSeconds": timeout_seconds,
                "listOptions.limit": limit,
                "listOptions.continue": continue_,
                "listOptions.sendInitialEvents": send_initial_events,
            },
            headers={"Authorization": self.token},
            data=None,
            verify=self.verify_ssl,
            cert=self.client_certs,
        )

        if resp.ok:
            return EventSourceWatchEvent(**resp.json())

        raise exception_from_server_response(resp)

    def event_sources_logs(
        self,
        namespace: Optional[str] = None,
        name: Optional[str] = None,
        event_source_type: Optional[str] = None,
        event_name: Optional[str] = None,
        grep: Optional[str] = None,
        container: Optional[str] = None,
        follow: Optional[bool] = None,
        previous: Optional[bool] = None,
        since_seconds: Optional[str] = None,
        seconds: Optional[str] = None,
        nanos: Optional[int] = None,
        timestamps: Optional[bool] = None,
        tail_lines: Optional[str] = None,
        limit_bytes: Optional[str] = None,
        insecure_skip_tls_verify_backend: Optional[bool] = None,
    ) -> EventsourceLogEntry:
        """API documentation."""
        assert valid_host_scheme(self.host), "The host scheme is required for service usage"
        resp = self._request(
            method="get",
            url=urljoin(self.host, "api/v1/stream/event-sources/{namespace}/logs").format(
                namespace=namespace if namespace is not None else self.namespace
            ),
            params={
                "name": name,
                "eventSourceType": event_source_type,
                "eventName": event_name,
                "grep": grep,
                "podLogOptions.container": container,
                "podLogOptions.follow": follow,
                "podLogOptions.previous": previous,
                "podLogOptions.sinceSeconds": since_seconds,
                "podLogOptions.sinceTime.seconds": seconds,
                "podLogOptions.sinceTime.nanos": nanos,
                "podLogOptions.timestamps": timestamps,
                "podLogOptions.tailLines": tail_lines,
                "podLogOptions.limitBytes": limit_bytes,
                "podLogOptions.insecureSkipTLSVerifyBackend": insecure_skip_tls_verify_backend,
            },
            headers={"Authorization": self.token},
            data=None,
            verify=self.verify_ssl,
            cert=self.client_certs,
        )

        if resp.ok:
            return EventsourceLogEntry(**resp.json())

        raise exception_from_server_response(resp)

    def watch_events(
        self,
        namespace: Optional[str] = None,
        label_selector: Optional[str] = None,
        field_selector: Optional[str] = None,
        watch: Optional[bool] = None,
        allow_watch_bookmarks: Optional[bool] = None,
        resource_version: Optional[str] = None,
        resource_version_match: Optional[str] = None,
        timeout_seconds: Optional[str] = None,
        limit: Optional[str] = None,
        continue_: Optional[str] = None,
        send_initial_events: Optional[bool] = None,
    ) -> Event:
        """API documentation."""
        assert valid_host_scheme(self.host), "The host scheme is required for service usage"
        resp = self._request(
            method="get",
            url=urljoin(self.host, "api/v1/stream/events/{namespace}").format(
                namespace=namespace if namespace is not None else self.namespace
            ),
            params={
                "listOptions.labelSelector": label_selector,
                "listOptions.fieldSelector": field_selector,
                "listOptions.watch": watch,
                "listOptions.allowWatchBookmarks": allow_watch_bookmarks,
                "listOptions.resourceVersion": resource_version,
                "listOptions.resourceVersionMatch": resource_version_match,
                "listOptions.timeoutSeconds": timeout_seconds,
                "listOptions.limit": limit,
                "listOptions.continue": continue_,
                "listOptions.sendInitialEvents": send_initial_events,
            },
            headers={"Authorization": self.token},
            data=None,
            verify=self.verify_ssl,
            cert=self.client_certs,
        )

        if resp.ok:
            return Event(**resp.json())

        raise exception_from_server_response(resp)

    def watch_sensors(
        self,
        namespace: Optional[str] = None,
        label_selector: Optional[str] = None,
        field_selector: Optional[str] = None,
        watch: Optional[bool] = None,
        allow_watch_bookmarks: Optional[bool] = None,
        resource_version: Optional[str] = None,
        resource_version_match: Optional[str] = None,
        timeout_seconds: Optional[str] = None,
        limit: Optional[str] = None,
        continue_: Optional[str] = None,
        send_initial_events: Optional[bool] = None,
    ) -> SensorWatchEvent:
        """API documentation."""
        assert valid_host_scheme(self.host), "The host scheme is required for service usage"
        resp = self._request(
            method="get",
            url=urljoin(self.host, "api/v1/stream/sensors/{namespace}").format(
                namespace=namespace if namespace is not None else self.namespace
            ),
            params={
                "listOptions.labelSelector": label_selector,
                "listOptions.fieldSelector": field_selector,
                "listOptions.watch": watch,
                "listOptions.allowWatchBookmarks": allow_watch_bookmarks,
                "listOptions.resourceVersion": resource_version,
                "listOptions.resourceVersionMatch": resource_version_match,
                "listOptions.timeoutSeconds": timeout_seconds,
                "listOptions.limit": limit,
                "listOptions.continue": continue_,
                "listOptions.sendInitialEvents": send_initial_events,
            },
            headers={"Authorization": self.token},
            data=None,
            verify=self.verify_ssl,
            cert=self.client_certs,
        )

        if resp.ok:
            return SensorWatchEvent(**resp.json())

        raise exception_from_server_response(resp)

    def sensors_logs(
        self,
        namespace: Optional[str] = None,
        name: Optional[str] = None,
        trigger_name: Optional[str] = None,
        grep: Optional[str] = None,
        container: Optional[str] = None,
        follow: Optional[bool] = None,
        previous: Optional[bool] = None,
        since_seconds: Optional[str] = None,
        seconds: Optional[str] = None,
        nanos: Optional[int] = None,
        timestamps: Optional[bool] = None,
        tail_lines: Optional[str] = None,
        limit_bytes: Optional[str] = None,
        insecure_skip_tls_verify_backend: Optional[bool] = None,
    ) -> SensorLogEntry:
        """API documentation."""
        assert valid_host_scheme(self.host), "The host scheme is required for service usage"
        resp = self._request(
            method="get",
            url=urljoin(self.host, "api/v1/stream/sensors/{namespace}/logs").format(
                namespace=namespace if namespace is not None else self.namespace
            ),
            params={
                "name": name,
                "triggerName": trigger_name,
                "grep": grep,
                "podLogOptions.container": container,
                "podLogOptions.follow": follow,
                "podLogOptions.previous": previous,
                "podLogOptions.sinceSeconds": since_seconds,
                "podLogOptions.sinceTime.seconds": seconds,
                "podLogOptions.sinceTime.nanos": nanos,
                "podLogOptions.timestamps": timestamps,
                "podLogOptions.tailLines": tail_lines,
                "podLogOptions.limitBytes": limit_bytes,
                "podLogOptions.insecureSkipTLSVerifyBackend": insecure_skip_tls_verify_backend,
            },
            headers={"Authorization": self.token},
            data=None,
            verify=self.verify_ssl,
            cert=self.client_certs,
        )

        if resp.ok:
            return SensorLogEntry(**resp.json())

        raise exception_from_server_response(resp)

    def get_user_info(self) -> GetUserInfoResponse:
        """API documentation."""
        assert valid_host_scheme(self.host), "The host scheme is required for service usage"
        resp = self._request(
            method="get",
            url=urljoin(self.host, "api/v1/userinfo"),
            params=None,
            headers={"Authorization": self.token},
            data=None,
            verify=self.verify_ssl,
            cert=self.client_certs,
        )

        if resp.ok:
            return GetUserInfoResponse(**resp.json())

        raise exception_from_server_response(resp)

    def get_version(self) -> Version:
        """API documentation."""
        assert valid_host_scheme(self.host), "The host scheme is required for service usage"
        resp = self._request(
            method="get",
            url=urljoin(self.host, "api/v1/version"),
            params=None,
            headers={"Authorization": self.token},
            data=None,
            verify=self.verify_ssl,
            cert=self.client_certs,
        )

        if resp.ok:
            return Version(**resp.json())

        raise exception_from_server_response(resp)

    def get_artifact_file(
        self,
        id_discriminator: str,
        id_: str,
        node_id: str,
        artifact_name: str,
        artifact_discriminator: str,
        namespace: Optional[str] = None,
    ) -> str:
        """Get an artifact."""
        assert valid_host_scheme(self.host), "The host scheme is required for service usage"
        resp = self._request(
            method="get",
            url=urljoin(
                self.host,
                "artifact-files/{namespace}/{idDiscriminator}/{id}/{nodeId}/{artifactDiscriminator}/{artifactName}",
            ).format(
                idDiscriminator=id_discriminator,
                id=id_,
                nodeId=node_id,
                artifactName=artifact_name,
                artifactDiscriminator=artifact_discriminator,
                namespace=namespace if namespace is not None else self.namespace,
            ),
            params=None,
            headers={"Authorization": self.token},
            data=None,
            verify=self.verify_ssl,
            cert=self.client_certs,
        )

        if resp.ok:
            return str(resp.content)

        raise exception_from_server_response(resp)

    def get_output_artifact_by_uid(self, uid: str, node_id: str, artifact_name: str) -> str:
        """Get an output artifact by UID."""
        assert valid_host_scheme(self.host), "The host scheme is required for service usage"
        resp = self._request(
            method="get",
            url=urljoin(self.host, "artifacts-by-uid/{uid}/{nodeId}/{artifactName}").format(
                uid=uid, nodeId=node_id, artifactName=artifact_name
            ),
            params=None,
            headers={"Authorization": self.token},
            data=None,
            verify=self.verify_ssl,
            cert=self.client_certs,
        )

        if resp.ok:
            return str(resp.content)

        raise exception_from_server_response(resp)

    def get_output_artifact(self, name: str, node_id: str, artifact_name: str, namespace: Optional[str] = None) -> str:
        """Get an output artifact."""
        assert valid_host_scheme(self.host), "The host scheme is required for service usage"
        resp = self._request(
            method="get",
            url=urljoin(self.host, "artifacts/{namespace}/{name}/{nodeId}/{artifactName}").format(
                name=name,
                nodeId=node_id,
                artifactName=artifact_name,
                namespace=namespace if namespace is not None else self.namespace,
            ),
            params=None,
            headers={"Authorization": self.token},
            data=None,
            verify=self.verify_ssl,
            cert=self.client_certs,
        )

        if resp.ok:
            return str(resp.content)

        raise exception_from_server_response(resp)

    def get_input_artifact_by_uid(self, uid: str, node_id: str, artifact_name: str) -> str:
        """Get an input artifact by UID."""
        assert valid_host_scheme(self.host), "The host scheme is required for service usage"
        resp = self._request(
            method="get",
            url=urljoin(self.host, "input-artifacts-by-uid/{uid}/{nodeId}/{artifactName}").format(
                uid=uid, nodeId=node_id, artifactName=artifact_name
            ),
            params=None,
            headers={"Authorization": self.token},
            data=None,
            verify=self.verify_ssl,
            cert=self.client_certs,
        )

        if resp.ok:
            return str(resp.content)

        raise exception_from_server_response(resp)

    def get_input_artifact(self, name: str, node_id: str, artifact_name: str, namespace: Optional[str] = None) -> str:
        """Get an input artifact."""
        assert valid_host_scheme(self.host), "The host scheme is required for service usage"
        resp = self._request(
            method="get",
            url=urljoin(self.host, "input-artifacts/{namespace}/{name}/{nodeId}/{artifactName}").format(
                name=name,
                nodeId=node_id,
                artifactName=artifact_name,
                namespace=namespace if namespace is not None else self.namespace,
            ),
            params=None,
            headers={"Authorization": self.token},
            data=None,
            verify=self.verify_ssl,
            cert=self.client_certs,
        )

        if resp.ok:
            return str(resp.content)

        raise exception_from_server_response(resp)


__all__ = ["EventsService"]



"""The task module provides the Task and TaskResult classes.

See https://argoproj.github.io/argo-workflows/walk-through/dag/
for more on using Tasks within a DAG.
"""

from __future__ import annotations

from enum import Enum
from typing import List, Optional, Union

from hera.workflows._mixins import (
    ArgumentsMixin,
    ItemMixin,
    ParameterMixin,
    SubNodeMixin,
    TemplateInvocatorSubNodeMixin,
    TemplateMixin,
)
from hera.workflows.models import (
    DAGTask as _ModelDAGTask,
    Template,
)
from hera.workflows.operator import Operator
from hera.workflows.protocol import Templatable
from hera.workflows.workflow_status import WorkflowStatus


class TaskResult(Enum):
    """The enumeration of Task Results.

    See Also:
        [Argo Depends Docs](https://argoproj.github.io/argo-workflows/enhanced-depends-logic/#depends)
    """

    failed = "Failed"
    succeeded = "Succeeded"
    errored = "Errored"
    skipped = "Skipped"
    omitted = "Omitted"
    daemoned = "Daemoned"
    any_succeeded = "AnySucceeded"
    all_failed = "AllFailed"


class Task(
    TemplateInvocatorSubNodeMixin,
    ArgumentsMixin,
    SubNodeMixin,
    ParameterMixin,
    ItemMixin,
):
    """Task is used to run a given template within a DAG. Must be instantiated under a DAG context."""

    dependencies: Optional[List[str]] = None
    depends: Optional[str] = None

    def _get_dependency_tasks(self) -> List[str]:
        if self.depends is None:
            return []

        # filter out operators
        all_operators = [o for o in Operator]
        tasks = [t for t in self.depends.split() if t not in all_operators]

        # remove dot suffixes
        task_names = [t.split(".")[0] for t in tasks]
        return task_names

    @property
    def _subtype(self) -> str:
        return "tasks"

    def next(self, other: Task, operator: Operator = Operator.and_, on: Optional[TaskResult] = None) -> Task:
        """Set self as a dependency of `other`."""
        assert issubclass(other.__class__, Task)

        condition = f".{on.value}" if on else ""

        if other.depends is None:
            # First dependency
            other.depends = self.name + condition
        elif self.name in other._get_dependency_tasks():
            raise ValueError(f"{self.name} already in {other.name}'s depends: {other.depends}")
        else:
            # Add follow-up dependency
            other.depends += f" {operator} {self.name + condition}"
        return other

    def __rrshift__(self, other: List[Union[Task, str]]) -> Task:
        """Set `other` as a dependency self."""
        assert isinstance(other, list), f"Unknown type {type(other)} specified using reverse right bitshift operator"
        for o in other:
            if isinstance(o, Task):
                o.next(self)
            else:
                assert isinstance(o, str), (
                    f"Unknown list item type {type(o)} specified using reverse right bitshift operator"
                )
                if self.depends is None:
                    self.depends = o
                else:
                    self.depends += f" && {o}"
        return self

    def __rshift__(self, other: Union[Task, List[Task]]) -> Union[Task, List[Task]]:
        """Set self as a dependency of `other` which can be a single Task or list of Tasks."""
        if isinstance(other, Task):
            return self.next(other)
        elif isinstance(other, list):
            for o in other:
                assert isinstance(o, Task), (
                    f"Unknown list item type {type(o)} specified using right bitshift operator `>>`"
                )
                self.next(o)
            return other
        raise ValueError(f"Unknown type {type(other)} provided to `__rshift__`")

    def __or__(self, other: Union[Task, str]) -> str:
        """Adds a condition of."""
        if isinstance(other, Task):
            return f"({self.name} || {other.name})"
        assert isinstance(other, str), f"Unknown type {type(other)} specified using `|` operator"
        return f"{self.name} || {other}"

    def on_workflow_status(self, status: WorkflowStatus, op: Operator = Operator.equals) -> Task:
        """Sets the current task to run when the workflow finishes with the specified status."""
        expression = f"{{{{workflow.status}}}} {op} {status}"
        if self.when:
            self.when += f" {Operator.and_} {expression}"
        else:
            self.when = expression
        return self

    def on_success(self, other: Task) -> Task:
        """Sets the current task to run when the given `other` task succeeds."""
        return self.next(other, on=TaskResult.succeeded)

    def on_failure(self, other: Task) -> Task:
        """Sets the current task to run when the given `other` task fails."""
        return self.next(other, on=TaskResult.failed)

    def on_error(self, other: Task) -> Task:
        """Sets the current task to run when the given `other` task errors."""
        return self.next(other, on=TaskResult.errored)

    def on_other_result(self, other: Task, value: str, operator: Operator = Operator.equals) -> Task:
        """Sets the current task to run when the given `other` task results in the specified `value` result."""
        expression = f"{other.result} {operator} {value}"
        if self.when:
            self.when += f" {Operator.and_} {expression}"
        else:
            self.when = expression
        other.next(self)
        return self

    def when_any_succeeded(self, other: Task) -> Task:
        """Sets the current task to run when the given `other` task succeedds."""
        assert (self.with_param is not None) or (self.with_sequence is not None), (
            "Can only use `when_all_failed` when using `with_param` or `with_sequence`"
        )

        return self.next(other, on=TaskResult.any_succeeded)

    def when_all_failed(self, other: Task) -> Task:
        """Sets the current task to run when the given `other` task has failed."""
        assert (self.with_param is not None) or (self.with_sequence is not None), (
            "Can only use `when_all_failed` when using `with_param` or `with_sequence`"
        )

        return self.next(other, on=TaskResult.all_failed)

    def _build_dag_task(self) -> _ModelDAGTask:
        _template = None
        if isinstance(self.template, str):
            _template = self.template
        elif isinstance(self.template, (Template, TemplateMixin)):
            _template = self.template.name

        _inline = None
        if isinstance(self.inline, Template):
            _inline = self.inline
        elif isinstance(self.inline, Templatable):
            _inline = self.inline._build_template()

        return _ModelDAGTask(
            arguments=self._build_arguments(),
            continue_on=self.continue_on,
            dependencies=self.dependencies,
            depends=self.depends,
            hooks=self.hooks,
            inline=_inline,
            name=self.name,
            on_exit=self._build_on_exit(),
            template=_template,
            template_ref=self.template_ref,
            when=self.when,
            with_items=self._build_with_items(),
            with_param=self._build_with_param(),
            with_sequence=self.with_sequence,
        )


__all__ = ["Task", "TaskResult"]



import sys
from typing import TYPE_CHECKING, Iterator, List, Optional, Tuple, Type, Union

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


from hera.shared._pydantic import _PYDANTIC_VERSION, FieldInfo, get_field_annotations, get_fields
from hera.shared._type_util import construct_io_from_annotation, get_workflow_annotation
from hera.shared.serialization import MISSING, serialize
from hera.workflows._context import _context
from hera.workflows.artifact import Artifact
from hera.workflows.models import (
    Arguments as ModelArguments,
    Artifact as ModelArtifact,
    Parameter as ModelParameter,
    ValueFrom,
)
from hera.workflows.parameter import Parameter

if _PYDANTIC_VERSION == 2:
    from pydantic import BaseModel as V2BaseModel
    from pydantic.v1 import BaseModel as V1BaseModel
    from pydantic_core import PydanticUndefined
else:
    from pydantic import BaseModel as V1BaseModel  # type: ignore[assignment]

    V2BaseModel = V1BaseModel  # type: ignore
    PydanticUndefined = None  # type: ignore[assignment]

if TYPE_CHECKING:
    # We add BaseModel as a parent class of the mixins only when type checking which allows it
    # to be used with either a V1 BaseModel or a V2 BaseModel
    from pydantic import BaseModel
else:
    # Subclassing `object` when using the real code (i.e. not type-checking) is basically a no-op
    BaseModel = object  # type: ignore


def _construct_io_from_fields(cls: Type[BaseModel]) -> Iterator[Tuple[str, FieldInfo, Union[Parameter, Artifact]]]:
    """Constructs a Parameter or Artifact object for all Pydantic fields based on their annotations.

    If a field has a Parameter or Artifact annotation, a copy will be returned, with missing
    fields filled out based on other metadata. Otherwise, a Parameter object will be constructed.
    """
    annotations = get_field_annotations(cls)
    for field, field_info in get_fields(cls).items():
        yield field, field_info, construct_io_from_annotation(field, annotations[field])


class InputMixin(BaseModel):
    def __new__(cls, **kwargs):
        if _context.declaring:
            # Intercept the declaration to avoid validation on the templated strings
            # We must then turn off declaring mode to be able to "construct" an instance
            # of the InputMixin subclass.
            _context.declaring = False
            instance = cls.construct(**kwargs)
            _context.declaring = True
            return instance
        else:
            return super(InputMixin, cls).__new__(cls)

    def __init__(self, /, **kwargs):
        if _context.declaring:
            # Return in order to skip validation of `construct`ed instance
            return

        super().__init__(**kwargs)

    @classmethod
    def _get_parameters(cls, object_override: Optional[Self] = None) -> List[Parameter]:
        parameters = []

        for field, field_info, param in _construct_io_from_fields(cls):
            if isinstance(param, Parameter):
                if param.default is not None:
                    raise ValueError(
                        "default cannot be set via the Parameter's default, use a Python default value instead."
                    )
                if object_override:
                    param.default = serialize(getattr(object_override, field))
                elif field_info.default is not None and field_info.default != PydanticUndefined:  # type: ignore
                    # Serialize the value (usually done in Parameter's validator)
                    param.default = serialize(field_info.default)  # type: ignore
                parameters.append(param)

        return parameters

    @classmethod
    def _get_artifacts(cls, add_missing_path: bool = False) -> List[Artifact]:
        artifacts = []

        for _, _, artifact in _construct_io_from_fields(cls):
            if isinstance(artifact, Artifact):
                if add_missing_path and artifact.path is None:
                    artifact.path = artifact._get_default_inputs_path()
                artifacts.append(artifact)
        return artifacts

    @classmethod
    def _get_inputs(cls, add_missing_path: bool = False) -> List[Union[Artifact, Parameter]]:
        return cls._get_artifacts(add_missing_path) + cls._get_parameters()

    @classmethod
    def _get_as_templated_arguments(cls) -> Self:
        """Returns the Input with templated values to propagate through a DAG/Steps function."""
        object_dict = {}

        for field, _, annotation in _construct_io_from_fields(cls):
            input_type = "parameters" if isinstance(annotation, Parameter) else "artifacts"
            object_dict[field] = "{{" + f"inputs.{input_type}.{annotation.name}" + "}}"

        return cls.construct(None, **object_dict)

    def _get_as_arguments(self) -> ModelArguments:
        params = []
        artifacts = []

        if isinstance(self, V1BaseModel):
            self_dict = self.dict()
        elif _PYDANTIC_VERSION == 2 and isinstance(self, V2BaseModel):
            self_dict = self.model_dump(warnings="none")

        for field, _, annotation in _construct_io_from_fields(type(self)):
            # The value may be a static value (of any time) if it has a default value, so we need to serialize it
            # If it is a templated string, it will be unaffected as `"{{mystr}}" == serialize("{{mystr}}")``
            templated_value = serialize(self_dict[field])
            name = annotation.name
            assert name is not None  # guaranteed by _get_workflow_annotations

            if isinstance(annotation, Parameter):
                params.append(ModelParameter(name=name, value=templated_value))
            else:
                artifacts.append(ModelArtifact(name=name, from_=templated_value))

        return ModelArguments(parameters=params or None, artifacts=artifacts or None)


class OutputMixin(BaseModel):
    def __new__(cls, **kwargs):
        if _context.declaring:
            # Intercept the declaration to avoid validation on the templated strings
            _context.declaring = False
            instance = cls.construct(**kwargs)
            _context.declaring = True
            return instance
        else:
            return super(OutputMixin, cls).__new__(cls)

    def __init__(self, /, **kwargs):
        if _context.declaring:
            # Return in order to skip validation of `construct`ed instance
            return

        super().__init__(**kwargs)

    @classmethod
    def _get_outputs(cls, add_missing_path: bool = False) -> List[Union[Artifact, Parameter]]:
        outputs: List[Union[Artifact, Parameter]] = []

        for field, field_info, annotation in _construct_io_from_fields(cls):
            if field in {"exit_code", "result"}:
                continue
            if isinstance(annotation, Parameter):
                if annotation.default is None:
                    default = field_info.default
                    if default is not None and default != PydanticUndefined:
                        annotation.default = serialize(default)

                if add_missing_path and (annotation.value_from is None or annotation.value_from.path is None):
                    annotation.value_from = ValueFrom(path=f"/tmp/hera-outputs/parameters/{annotation.name}")
            else:
                if add_missing_path and annotation.path is None:
                    annotation.path = f"/tmp/hera-outputs/artifacts/{annotation.name}"
            outputs.append(annotation)
        return outputs

    @classmethod
    def _get_output(cls, field_name: str) -> Union[Artifact, Parameter]:
        annotations = get_field_annotations(cls)
        annotation = annotations[field_name]
        if output := get_workflow_annotation(annotation):
            return output

        # Create a Parameter from basic type annotations
        default = get_fields(cls)[field_name].default
        if default is None or default == PydanticUndefined:
            default = MISSING
        return Parameter(name=field_name, default=default)  # type: ignore

    def _get_as_invocator_output(self) -> List[Union[Artifact, Parameter]]:
        """Get the Output model as the output of a dag/steps template.

        This lets dags and steps hoist task/step outputs into its own outputs.
        """
        outputs: List[Union[Artifact, Parameter]] = []

        if isinstance(self, V1BaseModel):
            self_dict = self.dict()
        elif _PYDANTIC_VERSION == 2 and isinstance(self, V2BaseModel):
            self_dict = self.model_dump(warnings="none")

        for field, _, annotation in _construct_io_from_fields(type(self)):
            if field in {"exit_code", "result"}:
                continue

            templated_value = self_dict[field]  # a string such as `"{{tasks.task_a.outputs.parameter.my_param}}"`

            if isinstance(annotation, Parameter):
                annotation.value_from = ValueFrom(parameter=templated_value)
            else:
                annotation.from_ = templated_value

            outputs.append(annotation)

        return outputs



"""The workflow module provides the Workflow class.

See https://argoproj.github.io/argo-workflows/workflow-concepts/#the-workflow
for more on Workflows.
"""

import logging
import time
from pathlib import Path
from typing import Annotated, Any, Callable, Dict, List, Optional, Type, TypeVar, Union, get_args

from typing_extensions import ParamSpec

from hera import _yaml
from hera.shared import global_config
from hera.shared._pydantic import BaseModel, validator
from hera.workflows._meta_mixins import HookMixin, ModelMapperMixin, TemplateDecoratorFuncsMixin
from hera.workflows._mixins import (
    ArgumentsMixin,
    ArgumentsT,
    MetricsMixin,
    MetricsT,
    VolumeMixin,
    VolumesT,
)
from hera.workflows.exceptions import InvalidType
from hera.workflows.models import (
    Affinity,
    ArtifactGC,
    ArtifactRepositoryRef,
    ExecutorConfig,
    HostAlias,
    LifecycleHook,
    LocalObjectReference,
    ManagedFieldsEntry,
    Metadata,
    ObjectMeta,
    OwnerReference,
    PersistentVolumeClaim,
    PodDisruptionBudgetSpec,
    PodDNSConfig,
    PodGC,
    PodSecurityContext,
    RetryStrategy,
    Synchronization,
    Template as _ModelTemplate,
    Time,
    Toleration,
    TTLStrategy,
    VolumeClaimGC,
    Workflow as _ModelWorkflow,
    WorkflowCreateRequest,
    WorkflowLintRequest,
    WorkflowMetadata,
    WorkflowSpec as _ModelWorkflowSpec,
    WorkflowStatus as _ModelWorkflowStatus,
    WorkflowTemplateRef,
)
from hera.workflows.parameter import Parameter
from hera.workflows.protocol import Templatable, TTemplate, TWorkflow, VolumeClaimable
from hera.workflows.service import WorkflowsService
from hera.workflows.template_set import TemplateSet
from hera.workflows.workflow_status import WorkflowStatus

ImagePullSecretsT = Optional[Union[LocalObjectReference, List[LocalObjectReference], str, List[str]]]

NAME_LIMIT = 63

T = TypeVar("T")
P = ParamSpec("P")


class _WorkflowModelMapper(ModelMapperMixin.ModelMapper):
    @classmethod
    def _get_model_class(cls) -> Type[BaseModel]:
        return _ModelWorkflow


class Workflow(
    ArgumentsMixin,
    HookMixin,
    VolumeMixin,
    MetricsMixin,
    ModelMapperMixin,
    TemplateDecoratorFuncsMixin,
):
    """The base Workflow class for Hera.

    Workflow implements the contextmanager interface so allows usage of `with`, under which
    any `hera.workflows.protocol.Templatable` object instantiated under the context will be
    added to the Workflow's list of templates.

    Workflows can be created directly on your Argo cluster via `create`. They can also be dumped
    to yaml via `to_yaml` or built according to the Argo schema via `build` to get an OpenAPI model
    object.
    """

    def _build_volume_claim_templates(self) -> Optional[List]:
        return ((self.volume_claim_templates or []) + (self._build_persistent_volume_claims() or [])) or None

    def _build_on_exit(self) -> Optional[str]:
        if isinstance(self.on_exit, Templatable):
            return self.on_exit._build_template().name  # type: ignore
        return self.on_exit

    def _build_templates(self) -> Optional[List[TTemplate]]:
        """Builds the templates into an Argo schema."""
        templates = []
        for template in self.templates:
            if isinstance(template, HookMixin):
                template = template._dispatch_hooks()

            if isinstance(template, Templatable):
                templates.append(template._build_template())
            elif isinstance(template, get_args(TTemplate)):
                templates.append(template)
            else:
                raise InvalidType(f"{type(template)} is not a valid template type")

            if isinstance(template, VolumeClaimable):
                claims = template._build_persistent_volume_claims()  # type: ignore
                # If there are no claims, continue, nothing to add
                if not claims:
                    continue
                # If there are no volume claim templates, set them to the constructed claims
                elif self.volume_claim_templates is None:
                    self.volume_claim_templates = claims
                else:
                    # otherwise, we need to merge the two lists of volume claim templates. This prioritizes the
                    # already existing volume claim templates under the assumption that the user has already set
                    # a claim template on the workflow intentionally, or the user is sharing the same volumes across
                    # different templates
                    current_volume_claims_map = {}
                    for claim in self.volume_claim_templates:
                        assert claim.metadata is not None, "expected a workflow volume claim with metadata"
                        assert claim.metadata.name is not None, "expected a named workflow volume claim"
                        current_volume_claims_map[claim.metadata.name] = claim

                    new_volume_claims_map = {}
                    for claim in claims:
                        assert claim.metadata is not None, "expected a volume claim with metadata"
                        assert claim.metadata.name is not None, "expected a named volume claim"
                        new_volume_claims_map[claim.metadata.name] = claim

                    for claim_name, claim in new_volume_claims_map.items():
                        if claim_name not in current_volume_claims_map:
                            self.volume_claim_templates.append(claim)
        return templates or None

    # Workflow fields - https://argoproj.github.io/argo-workflows/fields/#workflow
    api_version: Annotated[Optional[str], _WorkflowModelMapper("api_version")] = None
    kind: Annotated[Optional[str], _WorkflowModelMapper("kind")] = None
    status: Annotated[Optional[_ModelWorkflowStatus], _WorkflowModelMapper("status")] = None

    # ObjectMeta fields - https://argoproj.github.io/argo-workflows/fields/#objectmeta
    annotations: Annotated[Optional[Dict[str, str]], _WorkflowModelMapper("metadata.annotations")] = None
    creation_timestamp: Annotated[Optional[Time], _WorkflowModelMapper("metadata.creation_timestamp")] = None
    deletion_grace_period_seconds: Annotated[
        Optional[int], _WorkflowModelMapper("metadata.deletion_grace_period_seconds")
    ] = None
    deletion_timestamp: Annotated[Optional[Time], _WorkflowModelMapper("metadata.deletion_timestamp")] = None
    finalizers: Annotated[Optional[List[str]], _WorkflowModelMapper("metadata.finalizers")] = None
    generate_name: Annotated[Optional[str], _WorkflowModelMapper("metadata.generate_name")] = None
    generation: Annotated[Optional[int], _WorkflowModelMapper("metadata.generation")] = None
    labels: Annotated[Optional[Dict[str, str]], _WorkflowModelMapper("metadata.labels")] = None
    managed_fields: Annotated[Optional[List[ManagedFieldsEntry]], _WorkflowModelMapper("metadata.managed_fields")] = (
        None
    )
    name: Annotated[Optional[str], _WorkflowModelMapper("metadata.name")] = None
    namespace: Annotated[Optional[str], _WorkflowModelMapper("metadata.namespace")] = None
    owner_references: Annotated[Optional[List[OwnerReference]], _WorkflowModelMapper("metadata.owner_references")] = (
        None
    )
    resource_version: Annotated[Optional[str], _WorkflowModelMapper("metadata.resource_version")] = None
    self_link: Annotated[Optional[str], _WorkflowModelMapper("metadata.self_link")] = None
    uid: Annotated[Optional[str], _WorkflowModelMapper("metadata.uid")] = None

    # WorkflowSpec fields - https://argoproj.github.io/argo-workflows/fields/#workflowspec
    active_deadline_seconds: Annotated[Optional[int], _WorkflowModelMapper("spec.active_deadline_seconds")] = None
    affinity: Annotated[Optional[Affinity], _WorkflowModelMapper("spec.affinity")] = None
    archive_logs: Annotated[Optional[bool], _WorkflowModelMapper("spec.archive_logs")] = None
    artifact_gc: Annotated[Optional[ArtifactGC], _WorkflowModelMapper("spec.artifact_gc")] = None
    artifact_repository_ref: Annotated[
        Optional[ArtifactRepositoryRef], _WorkflowModelMapper("spec.artifact_repository_ref")
    ] = None
    automount_service_account_token: Annotated[
        Optional[bool], _WorkflowModelMapper("spec.automount_service_account_token")
    ] = None
    dns_config: Annotated[Optional[PodDNSConfig], _WorkflowModelMapper("spec.dns_config")] = None
    dns_policy: Annotated[Optional[str], _WorkflowModelMapper("spec.dns_policy")] = None
    entrypoint: Annotated[Optional[str], _WorkflowModelMapper("spec.entrypoint")] = None
    executor: Annotated[Optional[ExecutorConfig], _WorkflowModelMapper("spec.executor")] = None
    hooks: Annotated[Optional[Dict[str, LifecycleHook]], _WorkflowModelMapper("spec.hooks")] = None
    host_aliases: Annotated[Optional[List[HostAlias]], _WorkflowModelMapper("spec.host_aliases")] = None
    host_network: Annotated[Optional[bool], _WorkflowModelMapper("spec.host_network")] = None
    image_pull_secrets: Annotated[ImagePullSecretsT, _WorkflowModelMapper("spec.image_pull_secrets")] = None
    node_selector: Annotated[Optional[Dict[str, str]], _WorkflowModelMapper("spec.node_selector")] = None
    on_exit: Annotated[Optional[Union[str, Templatable]], _WorkflowModelMapper("spec.on_exit", _build_on_exit)] = None
    parallelism: Annotated[Optional[int], _WorkflowModelMapper("spec.parallelism")] = None
    pod_disruption_budget: Annotated[
        Optional[PodDisruptionBudgetSpec], _WorkflowModelMapper("spec.pod_disruption_budget")
    ] = None
    pod_gc: Annotated[Optional[PodGC], _WorkflowModelMapper("spec.pod_gc")] = None
    pod_metadata: Annotated[Optional[Metadata], _WorkflowModelMapper("spec.pod_metadata")] = None
    pod_priority: Annotated[Optional[int], _WorkflowModelMapper("spec.pod_priority")] = None
    pod_priority_class_name: Annotated[Optional[str], _WorkflowModelMapper("spec.pod_priority_class_name")] = None
    pod_spec_patch: Annotated[Optional[str], _WorkflowModelMapper("spec.pod_spec_patch")] = None
    priority: Annotated[Optional[int], _WorkflowModelMapper("spec.priority")] = None
    retry_strategy: Annotated[Optional[RetryStrategy], _WorkflowModelMapper("spec.retry_strategy")] = None
    scheduler_name: Annotated[Optional[str], _WorkflowModelMapper("spec.scheduler_name")] = None
    security_context: Annotated[Optional[PodSecurityContext], _WorkflowModelMapper("spec.security_context")] = None
    service_account_name: Annotated[Optional[str], _WorkflowModelMapper("spec.service_account_name")] = None
    shutdown: Annotated[Optional[str], _WorkflowModelMapper("spec.shutdown")] = None
    suspend: Annotated[Optional[bool], _WorkflowModelMapper("spec.suspend")] = None
    synchronization: Annotated[Optional[Synchronization], _WorkflowModelMapper("spec.synchronization")] = None
    template_defaults: Annotated[Optional[_ModelTemplate], _WorkflowModelMapper("spec.template_defaults")] = None
    templates: Annotated[
        List[Union[_ModelTemplate, Templatable]], _WorkflowModelMapper("spec.templates", _build_templates)
    ] = []
    tolerations: Annotated[Optional[List[Toleration]], _WorkflowModelMapper("spec.tolerations")] = None
    ttl_strategy: Annotated[Optional[TTLStrategy], _WorkflowModelMapper("spec.ttl_strategy")] = None
    volume_claim_gc: Annotated[Optional[VolumeClaimGC], _WorkflowModelMapper("spec.volume_claim_gc")] = None
    volume_claim_templates: Annotated[
        Optional[List[PersistentVolumeClaim]],
        _WorkflowModelMapper("spec.volume_claim_templates", _build_volume_claim_templates),
    ] = None
    workflow_metadata: Annotated[Optional[WorkflowMetadata], _WorkflowModelMapper("spec.workflow_metadata")] = None
    workflow_template_ref: Annotated[
        Optional[WorkflowTemplateRef], _WorkflowModelMapper("spec.workflow_template_ref")
    ] = None

    # Override types for mixin fields
    arguments: Annotated[
        ArgumentsT,
        _WorkflowModelMapper("spec.arguments", ArgumentsMixin._build_arguments),
    ] = None
    metrics: Annotated[
        MetricsT,
        _WorkflowModelMapper("spec.metrics", MetricsMixin._build_metrics),
    ] = None
    volumes: Annotated[VolumesT, _WorkflowModelMapper("spec.volumes", VolumeMixin._build_volumes)] = None

    # Hera-specific fields
    workflows_service: Optional[WorkflowsService] = None

    @validator("name", pre=True, always=True)
    def _set_name(cls, v):
        if v is not None and len(v) > NAME_LIMIT:
            raise ValueError(f"name must be no more than {NAME_LIMIT} characters: {v}")
        return v

    @validator("generate_name", pre=True, always=True)
    def _set_generate_name(cls, v):
        if v is not None and len(v) > NAME_LIMIT:
            raise ValueError(f"generate_name must be no more than {NAME_LIMIT} characters: {v}")
        return v

    @validator("api_version", pre=True, always=True)
    def _set_api_version(cls, v):
        if v is None:
            return global_config.api_version
        return v

    @validator("workflows_service", pre=True, always=True)
    def _set_workflows_service(cls, v):
        if v is None:
            return WorkflowsService()
        return v

    @validator("kind", pre=True, always=True)
    def _set_kind(cls, v):
        if v is None:
            return cls.__name__  # type: ignore
        return v

    @validator("namespace", pre=True, always=True)
    def _set_namespace(cls, v):
        if v is None:
            return global_config.namespace
        return v

    @validator("service_account_name", pre=True, always=True)
    def _set_service_account_name(cls, v):
        if v is None:
            return global_config.service_account_name
        return v

    @validator("image_pull_secrets", pre=True, always=True)
    def _set_image_pull_secrets(cls, v):
        if v is None:
            return None

        if isinstance(v, str):
            return [LocalObjectReference(name=v)]
        elif isinstance(v, LocalObjectReference):
            return [v]

        assert isinstance(v, list), (
            "`image_pull_secrets` expected to be either a `str`, a `LocalObjectReferences`, a list of `str`, "
            "or a list of `LocalObjectReferences`"
        )
        result = []
        for secret in v:
            if isinstance(secret, str):
                result.append(LocalObjectReference(name=secret))
            elif isinstance(secret, LocalObjectReference):
                result.append(secret)
        return result

    def get_parameter(self, name: str) -> Parameter:
        """Attempts to find and return a `Parameter` of the specified name."""
        arguments = self._build_arguments()
        if arguments is None:
            raise KeyError("Workflow has no arguments set")
        if arguments.parameters is None:
            raise KeyError("Workflow has no argument parameters set")

        parameters = arguments.parameters
        if next((p for p in parameters if p.name == name), None) is None:
            raise KeyError(f"`{name}` is not a valid workflow parameter")
        return Parameter(name=name, value=f"{{{{workflow.parameters.{name}}}}}")

    def build(self) -> TWorkflow:
        """Builds the Workflow and its components into an Argo schema Workflow object."""
        self = self._dispatch_hooks()

        model_workflow = _ModelWorkflow(
            metadata=ObjectMeta(),
            spec=_ModelWorkflowSpec(),
        )
        return _WorkflowModelMapper.build_model(Workflow, self, model_workflow)

    def to_dict(self) -> Any:
        """Builds the Workflow as an Argo schema Workflow object and returns it as a dictionary."""
        return self.build().dict(exclude_none=True, by_alias=True)

    def __eq__(self, other) -> bool:
        """Verifies equality of `self` with the specified `other`."""
        if other.__class__ is self.__class__:
            return self.to_dict() == other.to_dict()

        return False

    def to_yaml(self, *args, **kwargs) -> str:
        """Builds the Workflow as an Argo schema Workflow object and returns it as yaml string."""

        def human_readable_ordering(kv: tuple) -> int:
            """Key ordering function for ordering in a more human-readable fashion.

            Ordering is:
            1. "name" keys always first (if present)
            2. Primitives (not dicts/lists)
            3. lists
            4. dict
            """
            k, v = kv
            if k == "name" and isinstance(v, str):
                return 0
            if not isinstance(v, (dict, list)):
                return 1
            if isinstance(v, list):
                return 2
            return 3

        def order_dict(d: dict) -> dict[str, Any]:
            """Recursively orders `d` by the custom_ordering function by inserting them into a copy of the dict in order."""
            d_copy: dict[str, Any] = dict()
            for k, v in sorted(d.items(), key=lambda x: (human_readable_ordering(x), x[0])):
                if isinstance(v, dict):
                    d_copy[k] = order_dict(v)
                elif isinstance(v, list):
                    if v and isinstance(v[0], dict):
                        d_copy[k] = [order_dict(i) if isinstance(i, dict) else i for i in v]
                    elif v and isinstance(v[0], list):
                        d_copy[k] = [[order_dict(i) for i in sublist] for sublist in v]
                    else:
                        d_copy[k] = v
                else:
                    d_copy[k] = v
            return d_copy

        human_ordered_dict = order_dict(self.to_dict())
        return _yaml.dump(human_ordered_dict, *args, **kwargs)

    def create(self, wait: bool = False, poll_interval: int = 5) -> TWorkflow:
        """Creates the Workflow on the Argo cluster.

        Parameters
        ----------
        wait: bool = False
            If false then the workflow is created asynchronously and the function returns immediately.
            If true then the workflow is created and the function blocks until the workflow is done executing.
        poll_interval: int = 5
            The interval in seconds to poll the workflow status if wait is true. Ignored when wait is false.
        """
        assert self.workflows_service, "workflow service not initialized"
        assert self.namespace, "workflow namespace not defined"

        wf = self.workflows_service.create_workflow(
            WorkflowCreateRequest(workflow=self.build()),  # type: ignore
            namespace=self.namespace,
        )
        # set the workflow name to the name returned by the API, which helps cover the case of users relying on
        # `generate_name=True`
        self.name = wf.metadata.name

        if wait:
            return self.wait(poll_interval=poll_interval)
        return wf

    def wait(self, poll_interval: int = 5) -> TWorkflow:
        """Waits for the Workflow to complete execution.

        Parameters
        ----------
        poll_interval: int = 5
            The interval in seconds to poll the workflow status.
        """
        assert self.workflows_service is not None, "workflow service not initialized"
        assert self.namespace is not None, "workflow namespace not defined"
        assert self.name is not None, "workflow name not defined"

        # here we use the sleep interval to wait for the workflow post creation. This is to address a potential
        # race conditions such as:
        # 1. Argo server says "workflow was accepted" but the workflow is not yet created
        # 2. Hera wants to verify the status of the workflow, but it's not yet defined because it's not created
        # 3. Argo finally creates the workflow
        # 4. Hera throws an `AssertionError` because the phase assertion fails
        time.sleep(poll_interval)
        wf = self.workflows_service.get_workflow(self.name, namespace=self.namespace)
        assert wf.metadata.name is not None, f"workflow name not defined for workflow {self.name}"

        assert wf.status is not None, f"workflow status not defined for workflow {wf.metadata.name}"
        assert wf.status.phase is not None, f"workflow phase not defined for workflow status {wf.status}"
        status = WorkflowStatus.from_argo_status(wf.status.phase)

        # keep polling for workflow status until completed, at the interval dictated by the user
        while status == WorkflowStatus.running:
            time.sleep(poll_interval)
            wf = self.workflows_service.get_workflow(wf.metadata.name, namespace=self.namespace)
            assert wf.status is not None, f"workflow status not defined for workflow {wf.metadata.name}"
            assert wf.status.phase is not None, f"workflow phase not defined for workflow status {wf.status}"
            status = WorkflowStatus.from_argo_status(wf.status.phase)
        return wf

    def lint(self) -> TWorkflow:
        """Lints the Workflow using the Argo cluster."""
        assert self.workflows_service, "workflow service not initialized"
        assert self.namespace, "workflow namespace not defined"
        return self.workflows_service.lint_workflow(
            WorkflowLintRequest(workflow=self.build()),  # type: ignore
            namespace=self.namespace,
        )

    def _add_sub(self, node: Any):
        """Adds the given node (expected to satisfy the `Templatable` protocol) to the context."""
        if not isinstance(node, (Templatable, _ModelTemplate)):
            raise InvalidType(type(node))
        self.templates.append(node)

    def to_file(self, output_directory: Union[Path, str] = ".", name: str = "", *args, **kwargs) -> Path:
        """Writes the Workflow as an Argo schema Workflow object to a YAML file and returns the path to the file.

        Args:
            output_directory: The directory to write the file to. Defaults to the current working directory.
            name: The name of the file to write without the file extension.  Defaults to the Workflow's name or a
                  generated name.
            *args: Additional arguments to pass to `yaml.dump`.
            **kwargs: Additional keyword arguments to pass to `yaml.dump`.
        """
        workflow_name = self.name or (self.generate_name or "workflow").rstrip("-")
        name = name or workflow_name
        output_directory = Path(output_directory)
        output_path = Path(output_directory) / f"{name}.yaml"
        output_directory.mkdir(parents=True, exist_ok=True)
        output_path.write_text(self.to_yaml(*args, **kwargs))
        return output_path.absolute()

    @classmethod
    def from_dict(cls, model_dict: Dict) -> ModelMapperMixin:
        """Create a Workflow from a Workflow contained in a dict.

        Examples:
            >>> my_workflow = Workflow(name="my-workflow")
            >>> my_workflow == Workflow.from_dict(my_workflow.to_dict())
            True
        """
        return cls._from_dict(model_dict, _ModelWorkflow)

    @classmethod
    def from_yaml(cls, yaml_str: str) -> ModelMapperMixin:
        """Create a Workflow from a Workflow contained in a YAML string.

        Examples:
            >>> my_workflow = Workflow.from_yaml(yaml_str)
        """
        return cls._from_yaml(yaml_str, _ModelWorkflow)

    @classmethod
    def from_file(cls, yaml_file: Union[Path, str]) -> ModelMapperMixin:
        """Create a Workflow from a Workflow contained in a YAML file.

        Examples:
            >>> yaml_file = Path(...)
            >>> my_workflow = Workflow.from_file(yaml_file)
        """
        return cls._from_file(yaml_file, _ModelWorkflow)

    def get_workflow_link(self) -> str:
        """Returns the workflow link for the workflow."""
        assert self.workflows_service is not None, "Cannot fetch a workflow link without a service"
        assert self.name is not None, "Cannot fetch a workflow link without a workflow name"
        return self.workflows_service.get_workflow_link(self.name)

    def set_entrypoint(self, func: Callable[P, T]) -> Callable[P, T]:
        """Decorator function to set entrypoint."""
        if not hasattr(func, "template_name"):
            raise SyntaxError("`set_entrypoint` decorator must be above template decorator")

        if self.entrypoint is not None:
            if self.entrypoint == func.template_name:
                return func

            logging.warning(f"entrypoint is being reassigned from {self.entrypoint} to {func.template_name}")

        self.entrypoint = func.template_name  # type: ignore
        return func

    def add_template_set(self, template_set: TemplateSet) -> None:
        """Add the templates stored in the template_set to this Workflow."""
        for template in template_set.templates:
            self._add_sub(template)


__all__ = ["Workflow"]



"""[DO NOT EDIT MANUALLY] Auto-generated model classes.

Auto-generated by Hera via `make workflows-models`.
OpenAPI spec URL: https://raw.githubusercontent.com/argoproj/argo-workflows/v3.6.2/api/openapi-spec/swagger.json
"""

from hera.workflows.models.google.protobuf import Any
from hera.workflows.models.grpc.gateway.runtime import Error, StreamError
from hera.workflows.models.io.argoproj.workflow.v1alpha1 import (
    HTTP,
    Amount,
    ArchivedWorkflowDeletedResponse,
    ArchiveStrategy,
    Arguments,
    ArtGCStatus,
    Artifact,
    ArtifactGC,
    ArtifactLocation,
    ArtifactoryArtifact,
    ArtifactoryArtifactRepository,
    ArtifactPaths,
    ArtifactRepository,
    ArtifactRepositoryRef,
    ArtifactRepositoryRefStatus,
    AzureArtifact,
    AzureArtifactRepository,
    Backoff,
    BasicAuth,
    Cache,
    ClientCertAuth,
    ClusterWorkflowTemplate,
    ClusterWorkflowTemplateCreateRequest,
    ClusterWorkflowTemplateDeleteResponse,
    ClusterWorkflowTemplateLintRequest,
    ClusterWorkflowTemplateList,
    ClusterWorkflowTemplateUpdateRequest,
    CollectEventRequest,
    CollectEventResponse,
    Column,
    Condition,
    ContainerNode,
    ContainerSetRetryStrategy,
    ContainerSetTemplate,
    ContinueOn,
    Counter,
    CreateCronWorkflowRequest,
    CreateS3BucketOptions,
    CronWorkflow,
    CronWorkflowDeletedResponse,
    CronWorkflowList,
    CronWorkflowResumeRequest,
    CronWorkflowSpec,
    CronWorkflowStatus,
    CronWorkflowSuspendRequest,
    DAGTask,
    DAGTemplate,
    Data,
    DataSource,
    Event,
    EventResponse,
    ExecutorConfig,
    Gauge,
    GCSArtifact,
    GCSArtifactRepository,
    GetUserInfoResponse,
    GitArtifact,
    HDFSArtifact,
    HDFSArtifactRepository,
    Header,
    Histogram,
    HTTPArtifact,
    HTTPAuth,
    HTTPBodySource,
    HTTPHeader as V1alpha1HTTPHeader,
    HTTPHeaderSource,
    InfoResponse,
    Inputs,
    Item,
    LabelKeys,
    LabelValueFrom,
    LabelValues,
    LifecycleHook,
    Link,
    LintCronWorkflowRequest,
    LogEntry as V1alpha1LogEntry,
    ManifestFrom,
    MemoizationStatus,
    Memoize,
    Metadata,
    MetricLabel,
    Metrics,
    Mutex,
    MutexHolding,
    MutexStatus,
    NodeFlag,
    NodeStatus,
    NodeSynchronizationStatus,
    NoneStrategy,
    OAuth2Auth,
    OAuth2EndpointParam,
    OSSArtifact,
    OSSArtifactRepository,
    OSSLifecycleRule,
    Outputs,
    ParallelSteps,
    Parameter,
    Plugin,
    PodGC,
    Prometheus,
    RawArtifact,
    ResourceTemplate,
    ResubmitArchivedWorkflowRequest,
    RetryAffinity,
    RetryArchivedWorkflowRequest,
    RetryNodeAntiAffinity,
    RetryStrategy,
    S3Artifact,
    S3ArtifactRepository,
    S3EncryptionOptions,
    ScriptTemplate,
    SemaphoreHolding,
    SemaphoreRef,
    SemaphoreStatus,
    Sequence,
    StopStrategy,
    Submit,
    SubmitOpts,
    SuppliedValueFrom,
    SuspendTemplate,
    Synchronization,
    SynchronizationStatus,
    TarStrategy,
    Template,
    TemplateRef,
    TransformationStep,
    TTLStrategy,
    UpdateCronWorkflowRequest,
    UserContainer,
    ValueFrom,
    Version,
    VolumeClaimGC,
    Workflow,
    WorkflowCreateRequest,
    WorkflowDeleteResponse,
    WorkflowEventBinding,
    WorkflowEventBindingList,
    WorkflowEventBindingSpec,
    WorkflowLevelArtifactGC,
    WorkflowLintRequest,
    WorkflowList,
    WorkflowMetadata,
    WorkflowResubmitRequest,
    WorkflowResumeRequest,
    WorkflowRetryRequest,
    WorkflowSetRequest,
    WorkflowSpec,
    WorkflowStatus,
    WorkflowStep,
    WorkflowStopRequest,
    WorkflowSubmitRequest,
    WorkflowSuspendRequest,
    WorkflowTemplate,
    WorkflowTemplateCreateRequest,
    WorkflowTemplateDeleteResponse,
    WorkflowTemplateLintRequest,
    WorkflowTemplateList,
    WorkflowTemplateRef,
    WorkflowTemplateUpdateRequest,
    WorkflowTerminateRequest,
    WorkflowWatchEvent,
    ZipStrategy,
)
from hera.workflows.models.io.k8s.api.core.v1 import (
    Affinity,
    AppArmorProfile,
    AWSElasticBlockStoreVolumeSource,
    AzureDiskVolumeSource,
    AzureFileVolumeSource,
    Capabilities,
    CephFSVolumeSource,
    CinderVolumeSource,
    ClusterTrustBundleProjection,
    ConfigMapEnvSource,
    ConfigMapKeySelector,
    ConfigMapProjection,
    ConfigMapVolumeSource,
    Container,
    ContainerPort,
    ContainerResizePolicy,
    CSIVolumeSource,
    DownwardAPIProjection,
    DownwardAPIVolumeFile,
    DownwardAPIVolumeSource,
    EmptyDirVolumeSource,
    EnvFromSource,
    EnvVar,
    EnvVarSource,
    EphemeralVolumeSource,
    ExecAction,
    FCVolumeSource,
    FlexVolumeSource,
    FlockerVolumeSource,
    GCEPersistentDiskVolumeSource,
    GitRepoVolumeSource,
    GlusterfsVolumeSource,
    GRPCAction,
    HostAlias,
    HostPathVolumeSource,
    HTTPGetAction,
    HTTPHeader as V1HTTPHeader,
    ImagePullPolicy,
    ISCSIVolumeSource,
    KeyToPath,
    Lifecycle,
    LifecycleHandler,
    LocalObjectReference,
    ModifyVolumeStatus,
    NFSVolumeSource,
    NodeAffinity,
    NodeSelector,
    NodeSelectorRequirement,
    NodeSelectorTerm,
    ObjectFieldSelector,
    ObjectReference,
    PersistentVolumeClaim,
    PersistentVolumeClaimCondition,
    PersistentVolumeClaimSpec,
    PersistentVolumeClaimStatus,
    PersistentVolumeClaimTemplate,
    PersistentVolumeClaimVolumeSource,
    PhotonPersistentDiskVolumeSource,
    PodAffinity,
    PodAffinityTerm,
    PodAntiAffinity,
    PodDNSConfig,
    PodDNSConfigOption,
    PodSecurityContext,
    PortworxVolumeSource,
    PreferredSchedulingTerm,
    Probe,
    ProjectedVolumeSource,
    QuobyteVolumeSource,
    RBDVolumeSource,
    ResourceClaim,
    ResourceFieldSelector,
    ResourceRequirements,
    ScaleIOVolumeSource,
    SeccompProfile,
    SecretEnvSource,
    SecretKeySelector,
    SecretProjection,
    SecretVolumeSource,
    SecurityContext,
    SELinuxOptions,
    ServiceAccountTokenProjection,
    ServicePort,
    SleepAction,
    StorageOSVolumeSource,
    Sysctl,
    TCPSocketAction,
    Toleration,
    TypedLocalObjectReference,
    TypedObjectReference,
    Volume,
    VolumeDevice,
    VolumeMount,
    VolumeProjection,
    VolumeResourceRequirements,
    VsphereVirtualDiskVolumeSource,
    WeightedPodAffinityTerm,
    WindowsSecurityContextOptions,
)
from hera.workflows.models.io.k8s.api.policy.v1 import PodDisruptionBudgetSpec
from hera.workflows.models.io.k8s.apimachinery.pkg.api.resource import Quantity
from hera.workflows.models.io.k8s.apimachinery.pkg.apis.meta.v1 import (
    CreateOptions,
    FieldsV1,
    GroupVersionResource,
    LabelSelector,
    LabelSelectorRequirement,
    ListMeta,
    ManagedFieldsEntry,
    MicroTime,
    ObjectMeta,
    OwnerReference,
    Time,
)
from hera.workflows.models.io.k8s.apimachinery.pkg.util.intstr import IntOrString

__all__ = [
    "AWSElasticBlockStoreVolumeSource",
    "Affinity",
    "Amount",
    "Any",
    "AppArmorProfile",
    "ArchiveStrategy",
    "ArchivedWorkflowDeletedResponse",
    "Arguments",
    "ArtGCStatus",
    "Artifact",
    "ArtifactGC",
    "ArtifactLocation",
    "ArtifactPaths",
    "ArtifactRepository",
    "ArtifactRepositoryRef",
    "ArtifactRepositoryRefStatus",
    "ArtifactoryArtifact",
    "ArtifactoryArtifactRepository",
    "AzureArtifact",
    "AzureArtifactRepository",
    "AzureDiskVolumeSource",
    "AzureFileVolumeSource",
    "Backoff",
    "BasicAuth",
    "CSIVolumeSource",
    "Cache",
    "Capabilities",
    "CephFSVolumeSource",
    "CinderVolumeSource",
    "ClientCertAuth",
    "ClusterTrustBundleProjection",
    "ClusterWorkflowTemplate",
    "ClusterWorkflowTemplateCreateRequest",
    "ClusterWorkflowTemplateDeleteResponse",
    "ClusterWorkflowTemplateLintRequest",
    "ClusterWorkflowTemplateList",
    "ClusterWorkflowTemplateUpdateRequest",
    "CollectEventRequest",
    "CollectEventResponse",
    "Column",
    "Condition",
    "ConfigMapEnvSource",
    "ConfigMapKeySelector",
    "ConfigMapProjection",
    "ConfigMapVolumeSource",
    "Container",
    "ContainerNode",
    "ContainerPort",
    "ContainerResizePolicy",
    "ContainerSetRetryStrategy",
    "ContainerSetTemplate",
    "ContinueOn",
    "Counter",
    "CreateCronWorkflowRequest",
    "CreateOptions",
    "CreateS3BucketOptions",
    "CronWorkflow",
    "CronWorkflowDeletedResponse",
    "CronWorkflowList",
    "CronWorkflowResumeRequest",
    "CronWorkflowSpec",
    "CronWorkflowStatus",
    "CronWorkflowSuspendRequest",
    "DAGTask",
    "DAGTemplate",
    "Data",
    "DataSource",
    "DownwardAPIProjection",
    "DownwardAPIVolumeFile",
    "DownwardAPIVolumeSource",
    "EmptyDirVolumeSource",
    "EnvFromSource",
    "EnvVar",
    "EnvVarSource",
    "EphemeralVolumeSource",
    "Error",
    "Event",
    "EventResponse",
    "ExecAction",
    "ExecutorConfig",
    "FCVolumeSource",
    "FieldsV1",
    "FlexVolumeSource",
    "FlockerVolumeSource",
    "GCEPersistentDiskVolumeSource",
    "GCSArtifact",
    "GCSArtifactRepository",
    "GRPCAction",
    "Gauge",
    "GetUserInfoResponse",
    "GitArtifact",
    "GitRepoVolumeSource",
    "GlusterfsVolumeSource",
    "GroupVersionResource",
    "HDFSArtifact",
    "HDFSArtifactRepository",
    "HTTP",
    "HTTPArtifact",
    "HTTPAuth",
    "HTTPBodySource",
    "HTTPGetAction",
    "HTTPHeaderSource",
    "Header",
    "Histogram",
    "HostAlias",
    "HostPathVolumeSource",
    "ISCSIVolumeSource",
    "ImagePullPolicy",
    "InfoResponse",
    "Inputs",
    "IntOrString",
    "Item",
    "KeyToPath",
    "LabelKeys",
    "LabelSelector",
    "LabelSelectorRequirement",
    "LabelValueFrom",
    "LabelValues",
    "Lifecycle",
    "LifecycleHandler",
    "LifecycleHook",
    "Link",
    "LintCronWorkflowRequest",
    "ListMeta",
    "LocalObjectReference",
    "ManagedFieldsEntry",
    "ManifestFrom",
    "MemoizationStatus",
    "Memoize",
    "Metadata",
    "MetricLabel",
    "Metrics",
    "MicroTime",
    "ModifyVolumeStatus",
    "Mutex",
    "MutexHolding",
    "MutexStatus",
    "NFSVolumeSource",
    "NodeAffinity",
    "NodeFlag",
    "NodeSelector",
    "NodeSelectorRequirement",
    "NodeSelectorTerm",
    "NodeStatus",
    "NodeSynchronizationStatus",
    "NoneStrategy",
    "OAuth2Auth",
    "OAuth2EndpointParam",
    "OSSArtifact",
    "OSSArtifactRepository",
    "OSSLifecycleRule",
    "ObjectFieldSelector",
    "ObjectMeta",
    "ObjectReference",
    "Outputs",
    "OwnerReference",
    "ParallelSteps",
    "Parameter",
    "PersistentVolumeClaim",
    "PersistentVolumeClaimCondition",
    "PersistentVolumeClaimSpec",
    "PersistentVolumeClaimStatus",
    "PersistentVolumeClaimTemplate",
    "PersistentVolumeClaimVolumeSource",
    "PhotonPersistentDiskVolumeSource",
    "Plugin",
    "PodAffinity",
    "PodAffinityTerm",
    "PodAntiAffinity",
    "PodDNSConfig",
    "PodDNSConfigOption",
    "PodDisruptionBudgetSpec",
    "PodGC",
    "PodSecurityContext",
    "PortworxVolumeSource",
    "PreferredSchedulingTerm",
    "Probe",
    "ProjectedVolumeSource",
    "Prometheus",
    "Quantity",
    "QuobyteVolumeSource",
    "RBDVolumeSource",
    "RawArtifact",
    "ResourceClaim",
    "ResourceFieldSelector",
    "ResourceRequirements",
    "ResourceTemplate",
    "ResubmitArchivedWorkflowRequest",
    "RetryAffinity",
    "RetryArchivedWorkflowRequest",
    "RetryNodeAntiAffinity",
    "RetryStrategy",
    "S3Artifact",
    "S3ArtifactRepository",
    "S3EncryptionOptions",
    "SELinuxOptions",
    "ScaleIOVolumeSource",
    "ScriptTemplate",
    "SeccompProfile",
    "SecretEnvSource",
    "SecretKeySelector",
    "SecretProjection",
    "SecretVolumeSource",
    "SecurityContext",
    "SemaphoreHolding",
    "SemaphoreRef",
    "SemaphoreStatus",
    "Sequence",
    "ServiceAccountTokenProjection",
    "ServicePort",
    "SleepAction",
    "StopStrategy",
    "StorageOSVolumeSource",
    "StreamError",
    "Submit",
    "SubmitOpts",
    "SuppliedValueFrom",
    "SuspendTemplate",
    "Synchronization",
    "SynchronizationStatus",
    "Sysctl",
    "TCPSocketAction",
    "TTLStrategy",
    "TarStrategy",
    "Template",
    "TemplateRef",
    "Time",
    "Toleration",
    "TransformationStep",
    "TypedLocalObjectReference",
    "TypedObjectReference",
    "UpdateCronWorkflowRequest",
    "UserContainer",
    "V1HTTPHeader",
    "V1alpha1HTTPHeader",
    "V1alpha1LogEntry",
    "ValueFrom",
    "Version",
    "Volume",
    "VolumeClaimGC",
    "VolumeDevice",
    "VolumeMount",
    "VolumeProjection",
    "VolumeResourceRequirements",
    "VsphereVirtualDiskVolumeSource",
    "WeightedPodAffinityTerm",
    "WindowsSecurityContextOptions",
    "Workflow",
    "WorkflowCreateRequest",
    "WorkflowDeleteResponse",
    "WorkflowEventBinding",
    "WorkflowEventBindingList",
    "WorkflowEventBindingSpec",
    "WorkflowLevelArtifactGC",
    "WorkflowLintRequest",
    "WorkflowList",
    "WorkflowMetadata",
    "WorkflowResubmitRequest",
    "WorkflowResumeRequest",
    "WorkflowRetryRequest",
    "WorkflowSetRequest",
    "WorkflowSpec",
    "WorkflowStatus",
    "WorkflowStep",
    "WorkflowStopRequest",
    "WorkflowSubmitRequest",
    "WorkflowSuspendRequest",
    "WorkflowTemplate",
    "WorkflowTemplateCreateRequest",
    "WorkflowTemplateDeleteResponse",
    "WorkflowTemplateLintRequest",
    "WorkflowTemplateList",
    "WorkflowTemplateRef",
    "WorkflowTemplateUpdateRequest",
    "WorkflowTerminateRequest",
    "WorkflowWatchEvent",
    "ZipStrategy",
]



"""The suspend module provides the Suspend class.

The Suspend template in Hera provides a convenience wrapper around the "Intermediate Parameters" Argo feature.

See https://argoproj.github.io/argo-workflows/walk-through/suspending/ for more on suspending.

See https://argoproj.github.io/argo-workflows/intermediate-inputs/ for more on intermediate parameters.
"""

from typing import List, Optional, Union

from hera.workflows._meta_mixins import CallableTemplateMixin
from hera.workflows._mixins import IOMixin, TemplateMixin
from hera.workflows.models import (
    Inputs,
    Outputs,
    SuppliedValueFrom,
    SuspendTemplate as _ModelSuspendTemplate,
    Template as _ModelTemplate,
    ValueFrom,
)
from hera.workflows.parameter import Parameter


class Suspend(
    TemplateMixin,
    CallableTemplateMixin,
    IOMixin,
):
    """The Suspend template allows the user to pause a workflow for a specified length of time.

    The workflow can pause based on the given by `duration` or indefinitely (i.e. until manually resumed).
    The Suspend template also allows you to specify `intermediate_parameters` which will replicate the given
    parameters to the "inputs" and "outputs" of the template, resulting in a Suspend template that pauses and
    waits for values from the user for the given list of parameters.
    """

    duration: Optional[Union[int, str]] = None
    intermediate_parameters: List[Parameter] = []

    def _build_suspend_template(self) -> _ModelSuspendTemplate:
        return _ModelSuspendTemplate(
            duration=str(self.duration) if self.duration else None,
        )

    def _build_outputs(self) -> Optional[Outputs]:
        intermediate_param_outputs = []
        for param in self.intermediate_parameters:
            intermediate_param_outputs.append(
                Parameter(
                    name=param.name,
                    value_from=ValueFrom(supplied=SuppliedValueFrom()),
                    description=param.description,
                ).as_output()
            )

        all_outputs = super()._build_outputs()
        if all_outputs is not None:
            if all_outputs.parameters is None:
                all_outputs.parameters = []
            all_outputs.parameters.extend(intermediate_param_outputs)
            return all_outputs
        else:
            return Outputs(parameters=intermediate_param_outputs) if intermediate_param_outputs else None

    def _build_inputs(self) -> Optional[Inputs]:
        intermediate_param_inputs = []
        for param in self.intermediate_parameters:
            intermediate_param_inputs.append(param.as_input())

        all_inputs = super()._build_inputs()
        if all_inputs is not None:
            if all_inputs.parameters is None:
                all_inputs.parameters = []
            all_inputs.parameters.extend(intermediate_param_inputs)
            return all_inputs
        else:
            return Inputs(parameters=intermediate_param_inputs) if intermediate_param_inputs else None

    def _build_template(self) -> _ModelTemplate:
        return _ModelTemplate(
            active_deadline_seconds=self.active_deadline_seconds,
            affinity=self.affinity,
            archive_location=self.archive_location,
            automount_service_account_token=self.automount_service_account_token,
            executor=self.executor,
            fail_fast=self.fail_fast,
            host_aliases=self.host_aliases,
            init_containers=self._build_init_containers(),
            inputs=self._build_inputs(),
            memoize=self.memoize,
            metadata=self._build_metadata(),
            name=self.name,
            node_selector=self.node_selector,
            outputs=self._build_outputs(),
            plugin=self.plugin,
            priority_class_name=self.priority_class_name,
            priority=self.priority,
            retry_strategy=self.retry_strategy,
            scheduler_name=self.scheduler_name,
            security_context=self.pod_security_context,
            service_account_name=self.service_account_name,
            sidecars=self._build_sidecars(),
            suspend=self._build_suspend_template(),
            synchronization=self.synchronization,
            timeout=self.timeout,
            tolerations=self.tolerations,
        )


__all__ = ["Suspend"]
